# coding=utf-8
# Copyright 2021 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Plotting utils."""

from typing import Union

import jax.numpy as jnp
import matplotlib.pyplot as plt
from ott.tools import transport
import scipy


def bidimensional(x: jnp.ndarray, y: jnp.ndarray):
  """Applies PCA to reduce to bimensional data."""
  if x.shape[1] < 3:
    return x, y

  u, s, _ = scipy.sparse.linalg.svds(jnp.concatenate([x, y], axis=0), k=2)
  proj = u * s
  k = x.shape[0]
  return proj[:k], proj[k:]


def _couplings(ax,
               x: jnp.ndarray,
               y: jnp.ndarray,
               a: jnp.ndarray,
               b: jnp.ndarray,
               matrix: jnp.ndarray,
               threshold: float = 0.0,
               scale: int = 200,
               cmap: str = 'Purples'):
  """Plots 2-D couplings. Projects via PCA if data is higher dimensional."""
  x, y = bidimensional(x, y)

  sa, sb = jnp.min(a) / scale, jnp.min(b) / scale
  ax.scatter(*x.T, s=a / sa, edgecolors='k', marker='o', label='x')
  ax.scatter(*y.T, s=b / sb, edgecolors='k', marker='X', label='y')

  cmap = plt.get_cmap(cmap)
  u, v = jnp.where(matrix > threshold)
  c = matrix[jnp.where(matrix > threshold)]
  xy = jnp.concatenate([x[u], y[v]], axis=-1)
  for i in range(xy.shape[0]):
    strength = jnp.max(jnp.array(matrix.shape)) * c[i]
    ax.plot(xy[i, [0, 2]], xy[i, [1, 3]],
            linewidth=0.5 + 4 * strength,
            color=cmap(strength),
            zorder=0, alpha=0.7)
  ax.legend(fontsize=15)


def couplings(arg: Union[transport.Transport, jnp.ndarray],
              y: jnp.ndarray = None,
              a: jnp.ndarray = None,
              b: jnp.ndarray = None,
              matrix: jnp.ndarray = None,
              ax=None,
              **kwargs):
  """Plots 2D points and the couplings between them."""
  if ax is None:
    _, ax = plt.subplots(1, 1, figsize=(8, 5))

  if isinstance(arg, transport.Transport):
    ot = arg
    return _couplings(ax, ot.geom.x, ot.geom.y, ot.a, ot.b, ot.matrix, **kwargs)

  return _couplings(ax, arg, y, a, b, matrix, **kwargs)


def _barycenters(ax,
                 y: jnp.ndarray,
                 a: jnp.ndarray,
                 b: jnp.ndarray,
                 matrix: jnp.ndarray,
                 scale: int = 200):
  """Plots 2-D sinkhorn barycenters."""
  sa, sb = jnp.min(a) / scale, jnp.min(b) / scale
  ax.scatter(*y.T, s=b / sb, edgecolors='k', marker='X', label='y')
  tx = 1 / a[:, None] * jnp.matmul(matrix, y)
  ax.scatter(*tx.T, s=a / sa, edgecolors='k', marker='X', label='T(x)')
  ax.legend(fontsize=15)


def barycenters(arg: Union[transport.Transport, jnp.ndarray],
                a: jnp.ndarray = None,
                b: jnp.ndarray = None,
                matrix: jnp.ndarray = None,
                ax=None,
                **kwargs):
  """Plots the barycenters, from the Transport object or from arguments."""
  if ax is None:
    _, ax = plt.subplots(1, 1, figsize=(8, 5))

  if isinstance(arg, transport.Transport):
    ot = arg
    return _barycenters(ax, ot.geom.y, ot.a, ot.b, ot.matrix, **kwargs)

  return _barycenters(ax, arg, a, b, matrix, **kwargs)



# coding=utf-8
# Copyright 2021 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Several cost/norm functions for relevant vector types."""
import abc

import jax
from jax.lib import xla_bridge
import jax.numpy as jnp
from ott.geometry import matrix_square_root


def dot(x: jnp.ndarray, y: jnp.ndarray):
  """Accelerator dependent dot. Implemented to avoid OOMs with online mode."""
  platform = xla_bridge.get_backend().platform
  return jnp.sum(x * y) if platform == 'gpu' else jnp.vdot(x, y)


@jax.tree_util.register_pytree_node_class
class CostFn(abc.ABC):
  """A generic cost function, taking two vectors as input.

  Cost functions evaluate a function on a pair of inputs. For convenience,
  that function is split into two norms -- evaluated on each input separately --
  followed by a pairwise cost that involves both inputs, as in

  c(x,y) = norm(x) + norm(y) + pairwise(x,y)

  If the norm function is not implemented, that value is handled as a 0.
  """

  norm = None  #  no norm function created by default.

  @abc.abstractmethod
  def pairwise(self, x, y):
    pass

  def __call__(self, x, y):
    return self.pairwise(x, y) + (
        0 if self.norm is None else self.norm(x) + self.norm(y))  # pylint: disable=not-callable

  def all_pairs(self, x: jnp.ndarray, y: jnp.ndarray):
    """Computes matrix of all costs (including norms) for vectors in x / y.

    Args:
      x: [num_a, d] jnp.ndarray
      y: [num_b, d] jnp.ndarray
    Returns:
      [num_a, num_b] matrix of cost evaluations.
    """
    return jax.vmap(lambda x_: jax.vmap(lambda y_: self(x_, y_))(y))(x)

  def all_pairs_pairwise(self, x: jnp.ndarray, y: jnp.ndarray):
    """Computes matrix of all pairwise-costs (no norms) for vectors in x / y.

    Args:
      x: [num_a, d] jnp.ndarray
      y: [num_b, d] jnp.ndarray
    Returns:
      [num_a, num_b] matrix of pairwise cost evaluations.
    """
    return jax.vmap(lambda x_: jax.vmap(lambda y_: self.pairwise(x_, y_))(y))(x)

  def tree_flatten(self):
    return (), None

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    del aux_data
    return cls(*children)


@jax.tree_util.register_pytree_node_class
class Euclidean(CostFn):
  """Squared Euclidean distance CostFn."""

  def norm(self, x):
    return jnp.sum(x ** 2, axis=-1)

  def pairwise(self, x, y):
    return -2 * dot(x, y)


@jax.tree_util.register_pytree_node_class
class Cosine(CostFn):
  """Cosine distance CostFn."""

  def __init__(self, ridge=1e-8):
    super().__init__()
    self._ridge = ridge

  def pairwise(self, x, y):
    ridge = self._ridge
    x_norm = jnp.linalg.norm(x, axis=-1)
    y_norm = jnp.linalg.norm(y, axis=-1)
    cosine_similarity = dot(x, y) / (x_norm * y_norm + ridge)
    cosine_distance = 1.0 - cosine_similarity
    return cosine_distance


@jax.tree_util.register_pytree_node_class
class Bures(CostFn):
  """Bures distance between a pair of (mean, cov matrix) raveled as vectors."""

  def __init__(self, dimension, **kwargs):
    super().__init__()
    self._dimension = dimension
    self._sqrtm_kw = kwargs

  def norm(self, x):
    norm = jnp.sum(x[..., 0:self._dimension]**2, axis=-1)
    x_mat = jnp.reshape(x[..., self._dimension:],
                        (-1, self._dimension, self._dimension))

    norm += jnp.trace(x_mat, axis1=-2, axis2=-1)
    return norm

  def pairwise(self, x, y):
    mean_dot_prod = dot(x[0:self._dimension], y[0:self._dimension])
    x_mat = jnp.reshape(x[self._dimension:],
                        (self._dimension, self._dimension))
    y_mat = jnp.reshape(y[self._dimension:],
                        (self._dimension, self._dimension))

    sq_x = matrix_square_root.sqrtm(x_mat, self._dimension,
                                    **self._sqrtm_kw)[0]
    sq_x_y_sq_x = jnp.matmul(sq_x, jnp.matmul(y_mat, sq_x))
    sq__sq_x_y_sq_x = matrix_square_root.sqrtm(sq_x_y_sq_x, self._dimension,
                                               **self._sqrtm_kw)[0]
    return -2 * (
        mean_dot_prod + jnp.trace(sq__sq_x_y_sq_x, axis1=-2, axis2=-1))

  def tree_flatten(self):
    return (), (self._dimension, self._sqrtm_kw)

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    del children
    return cls(aux_data[0], **aux_data[1])


@jax.tree_util.register_pytree_node_class
class UnbalancedBures(CostFn):
  """Regularized, unbalanced Bures distance between triplets.

  This cost implements the value defined in https://arxiv.org/pdf/2006.02572.pdf
  Equation 37, 39, 40. We follow their notations. It is assumed inputs are given
  as triplets (mass, mean, covariance) raveled as vectors, in that order.
  """

  def __init__(self,
               dimension: int,
               gamma: float = 1,
               sigma: float = 1,
               **kwargs):
    super().__init__()
    self._dimension = dimension
    self._gamma = gamma
    self._sigma2 = sigma ** 2
    self._sqrtm_kw = kwargs

  def norm(self, x):
    return self._gamma * x[0]

  def pairwise(self, x, y):
    # Sets a few constants
    gam = self._gamma
    sig2 = self._sigma2
    lam = sig2 + gam / 2
    tau = gam / (2 * lam)

    # Extracts mass, mean vector, covariance matrices
    mass_x, mass_y = x[0], y[0]
    diff_means = x[1:1+self._dimension] - y[1:1+self._dimension]
    x_mat = jnp.reshape(x[1+self._dimension:],
                        (self._dimension, self._dimension))
    y_mat = jnp.reshape(y[1+self._dimension:],
                        (self._dimension, self._dimension))

    # Identity matrix of suitable size
    iden = jnp.eye(self._dimension)

    # Creates matrices needed in the computation
    tilde_a = 0.5 * gam * (iden - lam * jnp.linalg.inv(x_mat + lam * iden))
    tilde_b = 0.5 * gam * (iden - lam * jnp.linalg.inv(y_mat + lam * iden))

    tilde_a_b = jnp.matmul(tilde_a, tilde_b)
    c_mat = matrix_square_root.sqrtm(
        1 / tau * tilde_a_b + 0.25 * (sig2 ** 2) * iden, **self._sqrtm_kw)[0]
    c_mat -= 0.5 * sig2 * iden

    # Computes log determinants (their sign should be >0).
    sldet_c, ldet_c = jnp.linalg.slogdet(c_mat)
    sldet_t_ab, ldet_t_ab = jnp.linalg.slogdet(tilde_a_b)
    sldet_ab, ldet_ab = jnp.linalg.slogdet(jnp.matmul(x_mat, y_mat))
    sldet_c_ab, ldet_c_ab = jnp.linalg.slogdet(c_mat - 2 * tilde_a_b / gam)

    # Gathers all these results to compute log total mass of transport
    log_m_pi = (0.5 * self._dimension * sig2 / (gam + sig2)) * jnp.log(sig2)

    log_m_pi += (1 / (tau + 1)) * (jnp.log(mass_x) + jnp.log(mass_y) + ldet_c
                                   + 0.5 * (tau * ldet_t_ab - ldet_ab))

    log_m_pi += -jnp.sum(diff_means * jnp.linalg.solve(
        x_mat + y_mat + lam * iden, diff_means)) / (2 * (tau + 1))

    log_m_pi += - 0.5 * ldet_c_ab

    # If all logdet signs are 1, output value, nan otherwise.
    return jnp.where(
        sldet_c == 1 and sldet_c_ab == 1 and sldet_ab == 1 and sldet_t_ab == 1,
        2 * sig2 * mass_x * mass_y - 2 * (sig2 + gam) * jnp.exp(log_m_pi),
        jnp.nan)

  def tree_flatten(self):
    return (), (self._dimension, self._gamma, self._sigma2, self._sqrtm_kw)

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    del children
    return cls(aux_data[0], aux_data[1], aux_data[2], **aux_data[3])



# coding=utf-8
# Copyright 2021 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A Jax backprop friendly version of Matrix square root."""

import functools
from typing import Tuple

import jax
import jax.numpy as jnp
import numpy as np
from ott.core import fixed_point_loop


@functools.partial(jax.custom_vjp, nondiff_argnums=(1, 2, 3, 4, 5))
def sqrtm(
    x: jnp.ndarray,
    threshold: float = 1e-3,
    min_iterations: int = 0,
    inner_iterations: int = 10,
    max_iterations: int = 1000,
    regularization: float = 1e-3
):
  """Implements Higham algorithm to compute matrix square root of p.d. matrix.

  See reference below, eq. 2.6.b
  http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.6.8799&rep=rep1&type=pdf

  Args:
    x: a (batch of) square p.s.d. matrices of the same size.
    threshold: convergence tolerance threshold for Newton-Schulz iterations.
    min_iterations: min number of iterations after which error is computed.
    inner_iterations: error is re-evaluated every inner_iterations iterations.
    max_iterations: max number of iterations.
    regularization: small regularizer added to norm of x, before normalization.

  Returns:
    sqrt matrix of x (or x's if batch), its inverse, errors along iterates.
  """
  dimension = x.shape[-1]
  norm_x = jnp.linalg.norm(x, axis=(-2, -1)) * (1 + regularization)

  if jnp.ndim(x) > 2:
    norm_x = norm_x[..., jnp.newaxis, jnp.newaxis]

  def cond_fn(iteration, const, state):
    """Stopping criterion. Checking decrease of objective is needed here."""
    _, threshold = const
    errors, _, _ = state
    err = errors[iteration // inner_iterations-1]

    return jnp.logical_or(
        iteration == 0,
        jnp.logical_and(
            jnp.logical_and(jnp.isfinite(err), err > threshold),
            jnp.all(jnp.diff(errors) <= 0)))  # check decreasing obj, else stop

  def body_fn(iteration, const, state, compute_error):
    """Carries out matrix updates on y and z, stores error if requested.

    Args:
      iteration: iteration number
      const: tuple of constant parameters that do not change throughout the
        loop.
      state: state variables currently updated in the loop.
      compute_error: flag to indicate this iteration computes/stores an error

    Returns:
      state variables.
    """
    x, _ = const
    errors, y, z = state
    w = 0.5 * jnp.matmul(z, y)
    y = 1.5 * y - jnp.matmul(y, w)
    z = 1.5 * z - jnp.matmul(w, z)

    err = jnp.where(compute_error, new_err(x, norm_x, y), np.inf)

    errors = errors.at[iteration // inner_iterations].set(err)

    return errors, y, z

  def new_err(x, norm_x, y):
    res = x - norm_x * jnp.matmul(y, y)
    norm_fn = functools.partial(jnp.linalg.norm, axis=(-2, -1))
    return jnp.max(norm_fn(res) / norm_fn(x))

  y = x / norm_x
  z = jnp.eye(dimension)
  if jnp.ndim(x) > 2:
    z = jnp.tile(z, list(x.shape[:-2]) + [1, 1])
  errors = -jnp.ones((np.ceil(max_iterations / inner_iterations).astype(int),))
  state = (errors, y, z)
  const = (x, threshold)
  errors, y, z = fixed_point_loop.fixpoint_iter_backprop(
      cond_fn, body_fn, min_iterations, max_iterations, inner_iterations, const,
      state)
  sqrt_x = jnp.sqrt(norm_x) * y
  inv_sqrt_x = z / jnp.sqrt(norm_x)

  return sqrt_x, inv_sqrt_x, errors


def _solve_sylvester(
    a: jnp.ndarray,
    b: jnp.ndarray
) -> Tuple[jnp.ndarray]:
  """Solve a Sylvester equation of the form XA + AX = B for X.

  When we differentiate X^{1/2}X^{1/2} = X, we obtain the Sylvester equation
  d(X^{1/2}) X^{1/2} + X^{1/2} d(X^{1/2}) = dX. We solve below by reformulating
  as a single linear equation using the vec operator (see
  https://nhigham.com/2020/09/01/what-is-the-sylvester-equation/ )
  Note that this is quite computationally expensive (O(n^6)!).

  Similarly, when we differentiate X^{-1/2}X^{-1/2} = X^{-1}, we obtain the
  equation d(X^{-1/2}) X^{-1/2} + X^{-1/2} d(X^{-1/2}) = d X^{-1} =
  -x^{-T} dx x^{-T}

  Args:
    a: value A on the left hand side of the Sylvester equation
    b: value B on the right hand side of the Sylvester equation

  Returns:
    The solution X.
  """
  # TODO(geoffd): reimplement using the Bartels-Stewart algorithm to solve
  #   in O(n^3) time when there is a JAX solver.
  d = a.shape[-1]
  eye_d = jnp.eye(d)
  a_t = jnp.swapaxes(a, axis1=-2, axis2=-1)
  l = jnp.kron(a_t, eye_d) + jnp.kron(eye_d, a)
  b = b.reshape(list(b.shape[:-2]) + [d * d, 1])
  return jnp.linalg.solve(l, b).reshape(a.shape)


def sqrtm_fwd(
    x: jnp.ndarray,
    threshold: float,
    min_iterations: int,
    inner_iterations: int,
    max_iterations: int,
    regularization: float,
) -> Tuple[Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
           Tuple[jnp.ndarray, jnp.ndarray]]:
  """Forward pass of custom VJP."""
  sqrt_x, inv_sqrt_x, errors = sqrtm(
      x=x,
      threshold=threshold,
      min_iterations=min_iterations,
      inner_iterations=inner_iterations,
      max_iterations=max_iterations,
      regularization=regularization,
      )
  return ((sqrt_x, inv_sqrt_x, errors), (sqrt_x, inv_sqrt_x))


def sqrtm_bwd(
    threshold: float,
    min_iterations: int,
    inner_iterations: int,
    max_iterations: int,
    regularization: float,
    residual: Tuple[jnp.ndarray, jnp.ndarray],
    cotangent: Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
) -> Tuple[jnp.ndarray,]:
  """Compute the derivative by solving a Sylvester equation."""
  del threshold, min_iterations, inner_iterations, max_iterations, regularization
  sqrt_x, inv_sqrt_x = residual
  # ignores cotangent associated with errors
  cot_sqrt, cot_inv_sqrt, _ = cotangent

  # get d(x^{1/2})
  vjp_cot_sqrt = _solve_sylvester(a=sqrt_x, b=cot_sqrt)

  # If we differentiate x^{-1/2}x^{-1/2} = x^{-1}, we obtain a similar
  # Sylvester equation to the one used to obtain d(x^{1/2}) above:
  # d(x^{-1/2})x^{-1/2} + x^{-1/2}d(x^{-1/2}) = d(x^{-1})
  # We then use the fact that d(x^{-1}) = -x^{-T} dx x^{-T}
  # (See The Matrix Cookbook section on derivatives of an inverse
  # https://www.math.uwaterloo.ca/~hwolkowi/matrixcookbook.pdf )
  inv_x_t = jnp.swapaxes(
      jnp.matmul(inv_sqrt_x, inv_sqrt_x), axis1=-2, axis2=-1)
  vjp_cot_inv_sqrt = _solve_sylvester(
      a=inv_sqrt_x, b=-jnp.matmul(inv_x_t, jnp.matmul(cot_inv_sqrt, inv_x_t)))
  return (vjp_cot_sqrt + vjp_cot_inv_sqrt,)


sqrtm.defvjp(sqrtm_fwd, sqrtm_bwd)


# Specialized versions of sqrtm that compute only the square root or inverse.
# These functions have lower complexity gradients than sqrtm.


@jax.custom_vjp
def sqrtm_only(
    x: jnp.ndarray
) -> jnp.ndarray:
  return sqrtm(x)[0]


def sqrtm_only_fwd(
    x: jnp.ndarray
) -> Tuple[jnp.ndarray, jnp.ndarray]:
  sqrt_x = sqrtm(x)[0]
  return sqrt_x, sqrt_x


def sqrtm_only_bwd(
    sqrt_x: jnp.ndarray,
    cotangent: jnp.ndarray
) -> Tuple[jnp.ndarray]:
  vjp = _solve_sylvester(a=sqrt_x, b=cotangent)
  return (vjp,)


sqrtm_only.defvjp(sqrtm_only_fwd, sqrtm_only_bwd)


@jax.custom_vjp
def inv_sqrtm_only(
    x: jnp.ndarray,
) -> jnp.ndarray:
  return sqrtm(x)[1]


def inv_sqrtm_only_fwd(
    x: jnp.ndarray
) -> Tuple[jnp.ndarray, jnp.ndarray]:
  inv_sqrt_x = sqrtm(x)[1]
  return inv_sqrt_x, inv_sqrt_x


def inv_sqrtm_only_bwd(
    residual: Tuple[jnp.ndarray, jnp.ndarray],
    cotangent: jnp.ndarray) -> jnp.ndarray:
  inv_sqrt_x = residual
  inv_x_t = jnp.swapaxes(jnp.matmul(inv_sqrt_x, inv_sqrt_x), axis1=-2, axis2=-1)
  vjp = _solve_sylvester(
      a=inv_sqrt_x, b=-jnp.matmul(inv_x_t, jnp.matmul(cotangent, inv_x_t)))
  return (vjp,)


inv_sqrtm_only.defvjp(inv_sqrtm_only_fwd, inv_sqrtm_only_bwd)



# coding=utf-8
# Copyright 2021 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Implementation of Amos+(2017) input convex neural networks (ICNN)."""

import jax
import jax.numpy as jnp
from flax import linen as nn
from typing import Any, Callable, Sequence, Tuple

PRNGKey = Any
Shape = Tuple[int]
Dtype = Any  # this could be a real type?
Array = Any


class PositiveDense(nn.Module):
  """A linear transformation using a weight matrix with all entries positive.

  Args:
    dim_hidden: the number of output dim_hidden.
    beta: inverse temperature parameter of the softplus function (default: 1).
    use_bias: whether to add a bias to the output (default: True).
    dtype: the dtype of the computation (default: float32).
    precision: numerical precision of computation see `jax.lax.Precision`
      for details.
    kernel_init: initializer function for the weight matrix.
    bias_init: initializer function for the bias.
  """
  dim_hidden: int
  beta: float = 1.0
  use_bias: bool = True
  dtype: Any = jnp.float32
  precision: Any = None
  kernel_init: Callable[
    [PRNGKey, Shape, Dtype], Array] = nn.initializers.lecun_normal()
  bias_init: Callable[[PRNGKey, Shape, Dtype], Array] = nn.initializers.zeros

  @nn.compact
  def __call__(self, inputs):
    """Applies a linear transformation to inputs along the last dimension.

    Args:
      inputs: The nd-array to be transformed.
    Returns:
      The transformed input.
    """
    inputs = jnp.asarray(inputs, self.dtype)
    kernel = self.param(
      'kernel', self.kernel_init, (inputs.shape[-1], self.dim_hidden))
    scaled_kernel = self.beta * kernel
    kernel = jnp.asarray(
      1 / self.beta * nn.softplus(scaled_kernel), self.dtype)
    y = jax.lax.dot_general(
      inputs, kernel, (((inputs.ndim - 1,), (0,)), ((), ())),
      precision=self.precision)
    if self.use_bias:
      bias = self.param('bias', self.bias_init, (self.dim_hidden,))
      bias = jnp.asarray(bias, self.dtype)
      y = y + bias
    return y


class ICNN(nn.Module):
  """Input convex neural network (ICNN) architeture.

  Args:
    dim_hidden: sequence specifying size of hidden dimensions. The
      output dimension of the last layer is 1 by default.
    init_std: value of standard deviation of weight initialization method.
    init_fn: choice of initialization method for weight matrices (default:
      `jax.nn.initializers.normal`).
    act_fn: choice of activation function used in network architecture
      (needs to be convex, default: `nn.leaky_relu`).
  """

  dim_hidden: Sequence[int]
  init_std: float = 0.1
  init_fn: Callable = jax.nn.initializers.normal
  act_fn: Callable = nn.leaky_relu

  def setup(self):
    num_hidden = len(self.dim_hidden)

    w_zs = list()

    for i in range(1, num_hidden):
      w_zs.append(PositiveDense(
        self.dim_hidden[i], kernel_init=self.init_fn(self.init_std),
        use_bias=False))
    w_zs.append(PositiveDense(
      1, kernel_init=self.init_fn(self.init_std), use_bias=False))
    self.w_zs = w_zs

    w_xs = list()
    for i in range(num_hidden):
      w_xs.append(nn.Dense(
        self.dim_hidden[i], kernel_init=self.init_fn(self.init_std),
        use_bias=True))
    w_xs.append(nn.Dense(
      1, kernel_init=self.init_fn(self.init_std), use_bias=True))
    self.w_xs = w_xs

  @nn.compact
  def __call__(self, x):
    """Applies ICNN module.

    Args:
      x: jnp.ndarray<float>[batch_size, n_features]: input to the ICNN.

    Returns:
      jnp.ndarray<float>[1]: output of ICNN.
    """
    z = self.act_fn(self.w_xs[0](x))
    z = jnp.multiply(z, z)

    for Wz, Wx in zip(self.w_zs[:-1], self.w_xs[1:-1]):
      z = self.act_fn(jnp.add(Wz(z), Wx(x)))
    y = jnp.add(self.w_zs[-1](z), self.w_xs[-1](x))

    return jnp.squeeze(y)



# coding=utf-8
# Copyright 2021 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A Jax version of Sinkhorn's algorithm."""
import abc
import collections
import functools
from typing import Optional, Union, Tuple, Any, Dict

import jax
import jax.numpy as jnp
from ott.core import sinkhorn
from ott.geometry import costs
from ott.geometry import epsilon_scheduler
from ott.geometry import geometry
from ott.geometry import pointcloud

GromovWassersteinOutput = collections.namedtuple(
    'GromovWassersteinOutput', ['f', 'g', 'transport', 'cost_matrix', 'gw_cost',
                                'reg_gw_cost', 'reg_gw_cost_arr',
                                'errors_sinkhorn', 'converged_sinkhorn'])


@jax.tree_util.register_pytree_node_class
class GWLoss(abc.ABC):
  """A loss function for Gromov-Wasserstein.

  The loss can be written as:
  L(x, y) = fn_x(x) + fn_y(y) - left_x(x) * right_y(y)
  """

  @abc.abstractmethod
  def fn_x(self, x):
    pass

  @abc.abstractmethod
  def fn_y(self, y):
    pass

  @abc.abstractmethod
  def left_x(self, x):
    pass

  @abc.abstractmethod
  def right_y(self, y):
    pass

  def tree_flatten(self):
    return (), None

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    del aux_data
    return cls(*children)


@jax.tree_util.register_pytree_node_class
class GWSqEuclLoss(GWLoss):
  """Implements the squared Euclidean distance for the Gromov-Wasserstein loss.
  """

  def fn_x(self, x):
    return x ** 2.0

  def fn_y(self, y):
    return y ** 2.0

  def left_x(self, x):
    return x

  def right_y(self, y):
    return 2.0 * y


@jax.tree_util.register_pytree_node_class
class GWKlLoss(GWLoss):
  """Implements the KL divergence for the Gromov-Wasserstein loss."""

  def __init__(self, clipping_value=1e-8):
    self.clipping_value = clipping_value

  def fn_x(self, x):
    return -jax.scipy.special.entr(x) - x

  def fn_y(self, y):
    return y

  def left_x(self, x):
    return x

  def right_y(self, y):
    return jnp.log(jnp.clip(y, self.clipping_value))


_GW_LOSSES = [GWSqEuclLoss, GWKlLoss]
GW_LOSSES = {cls.__name__.strip('GW').strip('Loss').lower(): cls()
             for cls in _GW_LOSSES}


def gromov_wasserstein(
    geom_x: geometry.Geometry,
    geom_y: geometry.Geometry,
    a: Optional[jnp.ndarray] = None,
    b: Optional[jnp.ndarray] = None,
    epsilon: Union[epsilon_scheduler.Epsilon, float] = 1.,
    loss: Union[str, GWLoss] = 'sqeucl',
    max_iterations: int = 20,
    jit: bool = False,
    warm_start: bool = True,
    sinkhorn_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs) -> GromovWassersteinOutput:
  """Fits Gromov Wasserstein.

  Args:
    geom_x: a Geometry object for the first view.
    geom_y: a second Geometry object for the second view.
    a: jnp.ndarray<float>[num_a,] or jnp.ndarray<float>[batch,num_a] weights.
    b: jnp.ndarray<float>[num_b,] or jnp.ndarray<float>[batch,num_b] weights.
    epsilon: a regularization parameter or a epsilon_scheduler.Epsilon object.
    loss: str or GWLoss object, whether to use the square euclidean loss
     ('sqeucl' or GWSqEuclLoss), the KL loss ('kl' or GWKlLoss) or a
     user-defined instance of GWLoss.
    max_iterations: int32, the maximum number of outer iterations for
     Gromov Wasserstein.
    jit: bool, if True, jits the function.
    warm_start: bool, optional initialisation of the potentials/scalings w.r.t.
     first and second marginals between each call to sinkhorn
     (default is True).
    sinkhorn_kwargs: Optionally a dictionary containing the keywords arguments
     for calls to the sinkhorn function.
    **kwargs: additional kwargs for epsilon.

  Returns:
    a GromovWassersteinOutput named tuple.
  """
  loss_fn = GW_LOSSES.get(loss, None) if isinstance(loss, str) else loss
  if loss_fn is None:
    raise ValueError('Unknown loss. Either pass an instance of GWLoss or '
                     f'a string among: [{",".join(GW_LOSSES.keys())}]')
  sinkhorn_kwargs = {} if sinkhorn_kwargs is None else sinkhorn_kwargs

  tau_a = sinkhorn_kwargs.get('tau_a', 1.0)
  tau_b = sinkhorn_kwargs.get('tau_b', 1.0)
  if tau_a != 1.0 or tau_b != 1.0:
    raise ValueError('Unbalanced Gromov-Wasserstein is not supported yet.')

  num_a = geom_x.shape[0]
  num_b = geom_y.shape[0]
  a = jnp.ones((num_a,)) / num_a if a is None else a
  b = jnp.ones((num_b,)) / num_b if b is None else b

  gromov_partial = functools.partial(
      _gw_iterations, epsilon=epsilon, loss=loss_fn,
      max_iterations=max_iterations, warm_start=warm_start,
      sinkhorn_kwargs=sinkhorn_kwargs, **kwargs)
  gromov_fn = jax.jit(gromov_partial) if jit else gromov_partial
  (f, g, geom_gw, reg_gw_cost_arr, errors_sinkhorn,
      converged_sinkhorn) = gromov_fn(geom_x, geom_y, a, b)

  # TODO(lpapaxanthos): remove stop_gradient when using backprop
  geom_gw = _update_geometry_gw(
      jax.lax.stop_gradient(geom_gw),
      geom_x, geom_y,
      jax.lax.stop_gradient(f),
      jax.lax.stop_gradient(g),
      loss_fn, **kwargs)
  transport = geom_gw.transport_from_potentials(f, g)
  cost_matrix = 0.5 * geom_gw.cost_matrix
  gw_cost = jnp.sum(transport * cost_matrix)
  reg_gw_cost = sinkhorn.ent_reg_cost(
      geom_gw, a, b, tau_a, tau_b,
      jax.lax.stop_gradient(f),
      jax.lax.stop_gradient(g),
      lse_mode=True)
  return GromovWassersteinOutput(f, g, transport, cost_matrix, gw_cost,
                                 reg_gw_cost, reg_gw_cost_arr, errors_sinkhorn,
                                 converged_sinkhorn)


def _gw_iterations(
    geom_x: geometry.Geometry,
    geom_y: geometry.Geometry,
    a: jnp.ndarray,
    b: jnp.ndarray,
    epsilon: Union[epsilon_scheduler.Epsilon, float],
    loss: GWLoss,
    max_iterations: int,
    warm_start: bool,
    sinkhorn_kwargs: Optional[Dict[str, Any]],
    **kwargs) -> Tuple[jnp.ndarray, jnp.ndarray, geometry.Geometry, jnp.ndarray,
                       jnp.ndarray, jnp.ndarray]:
  """Fits Gromov Wasserstein.

  Args:
    geom_x: a Geometry object for the first view.
    geom_y: a second Geometry object for the second view.
    a: jnp.ndarray<float>[num_a,] or jnp.ndarray<float>[batch,num_a] weights.
    b: jnp.ndarray<float>[num_b,] or jnp.ndarray<float>[batch,num_b] weights.
    epsilon: a regularization parameter or a epsilon_scheduler.Epsilon object.
    loss: GWLoss object.
    max_iterations: int, the maximum number of outer iterations for
     Gromov Wasserstein.
    warm_start: bool, optional initialisation of the potentials/scalings w.r.t.
     first and second marginals between each call to sinkhorn.
    sinkhorn_kwargs: Optionally a dictionary containing the keywords arguments
     for calls to the sinkhorn function.
    **kwargs: additional kwargs for epsilon.

  Returns:
    f: potential.
    g: potential.
    geom_gw: a Geometry object for Gromov-Wasserstein (GW).
    reg_gw_cost_arr: ndarray of regularised GW costs.
    errors_sinkhorn: ndarray [max_iterations, p], where p depends on
     sinkhorn_kwargs, of errors for the Sinkhorn algorithm for each gromov
     iteration (axis 0) and regularly spaced sinkhorn iterations (axis 1).
    converged_sinkhorn: ndarray [max_iterations,] of flags indicating
    that the sinkhorn algorithm converged.
  """
  lse_mode = sinkhorn_kwargs.get('lse_mode', True)

  geom_gw = _init_geometry_gw(
      geom_x, geom_y, jax.lax.stop_gradient(a), jax.lax.stop_gradient(b),
      epsilon, loss, **kwargs)
  f, g, reg_gw_cost, errors_sinkhorn, converged_sinkhorn = sinkhorn.sinkhorn(
      geom_gw, a, b, **sinkhorn_kwargs)
  carry = geom_gw, f, g
  update_geom_partial = functools.partial(_update_geometry_gw, geom_x=geom_x,
                                          geom_y=geom_y, loss=loss, **kwargs)
  sinkhorn_partial = functools.partial(sinkhorn.sinkhorn, a=a, b=b,
                                       **sinkhorn_kwargs)
  def body_fn(carry=carry, x=None):
    del x
    geom_gw, f, g = carry
    geom_gw = update_geom_partial(geom=geom_gw, f=f, g=g)
    init_dual_a = ((f if lse_mode else geom_gw.scaling_from_potential(f))
                   if warm_start else None)
    init_dual_b = ((g if lse_mode else geom_gw.scaling_from_potential(g))
                   if warm_start else None)
    f, g, reg_gw_cost, errors_sinkhorn, converged_sinkhorn = sinkhorn_partial(
        geom=geom_gw, init_dual_a=init_dual_a, init_dual_b=init_dual_b)
    return (geom_gw, f, g), (reg_gw_cost, errors_sinkhorn, converged_sinkhorn)

  carry, out = jax.lax.scan(f=body_fn, init=carry, xs=None,
                            length=max_iterations - 1)

  geom_gw, f, g = carry
  reg_gw_cost_arr = jnp.concatenate((jnp.array([reg_gw_cost]), out[0]))
  errors_sinkhorn = jnp.concatenate((jnp.array([errors_sinkhorn]), out[1]))
  converged_sinkhorn = jnp.concatenate((jnp.array([converged_sinkhorn]),
                                        out[2]))
  return (f, g, geom_gw, reg_gw_cost_arr, errors_sinkhorn, converged_sinkhorn)


def _init_geometry_gw(
    geom_x: geometry.Geometry,
    geom_y: geometry.Geometry,
    a: jnp.ndarray,
    b: jnp.ndarray,
    epsilon: Union[epsilon_scheduler.Epsilon, float],
    loss: GWLoss,
    **kwargs) -> geometry.Geometry:
  """Initialises the cost matrix for the geometry object for GW.

  The equation follows Equation 6, Proposition 1 of
  http://proceedings.mlr.press/v48/peyre16.pdf.

  Args:
    geom_x: a Geometry object for the first view.
    geom_y: a second Geometry object for the second view.
    a: jnp.ndarray<float>[num_a,], weights.
    b: jnp.ndarray<float>[num_b,], weights.
    epsilon: a regularization parameter or a epsilon_scheduler.Epsilon object.
    loss: a GWLossFn object.
    **kwargs: additional kwargs to epsilon.

  Returns:
    A Geometry object for Gromov-Wasserstein.
  """
  # Initialization of the transport matrix in the balanced case, following
  # http://proceedings.mlr.press/v48/peyre16.pdf
  ab = a[:, None] * b[None, :]
  marginal_x = ab.sum(1)
  marginal_y = ab.sum(0)
  marginal_dep_term = _marginal_dependent_cost(
      marginal_x, marginal_y, geom_x, geom_y, loss)

  tmp = geom_x.apply_cost(ab, axis=1, fn=loss.left_x)
  cost_matrix = marginal_dep_term - geom_y.apply_cost(
      tmp.T, axis=1, fn=loss.right_y).T
  return geometry.Geometry(cost_matrix=cost_matrix,
                           epsilon=epsilon, **kwargs)


def _update_geometry_gw(
    geom: geometry.Geometry,
    geom_x: geometry.Geometry,
    geom_y: geometry.Geometry,
    f: jnp.ndarray,
    g: jnp.ndarray,
    loss: GWLoss,
    **kwargs) -> geometry.Geometry:
  """Updates the geometry object for GW by updating the cost matrix.

  The cost matrix equation follows Equation 6, Proposition 1 of
  http://proceedings.mlr.press/v48/peyre16.pdf.

  Let :math:`p` [num_a,] be the marginal of the transport matrix for samples
  from geom_x and :math:`q` [num_b,] be the marginal of the transport matrix for
  samples from geom_y. Let :math:`T` [num_a, num_b] be the transport matrix.
  The cost matrix equation can be written as:

  cost_matrix = marginal_dep_term
              + left_x(cost_x) :math:`T` right_y(cost_y):math:`^T`

  Args:
    geom: a Geometry object carrying the cost matrix of Gromov Wasserstein.
    geom_x: a Geometry object for the first view.
    geom_y: a second Geometry object for the second view.
    f: jnp.ndarray<float>[num_a,], potentials.
    g: jnp.ndarray<float>[num_b,], potentials.
    loss: a GWLossFn object.
    **kwargs: additional kwargs for epsilon.

  Returns:
    A Geometry object for Gromov-Wasserstein.
  """
  def apply_cost_fn(geom):
    condition = is_sqeuclidean(geom) and isinstance(loss, GWSqEuclLoss)
    return geom.vec_apply_cost if condition else geom.apply_cost

  def is_sqeuclidean(geom):
    return (isinstance(geom, pointcloud.PointCloud)
            and geom.power == 2.0
            and isinstance(geom._cost_fn, costs.Euclidean))

  def is_online(geom):
    return isinstance(geom, pointcloud.PointCloud) and geom._online

  # Computes tmp = cost_matrix_x * transport
  if is_online(geom_x) or is_sqeuclidean(geom_x):
    transport = geom.transport_from_potentials(f, g)
    tmp = apply_cost_fn(geom_x)(transport, axis=1,
                                fn=loss.left_x)
  else:
    tmp = geom.apply_transport_from_potentials(
        f, g, loss.left_x(geom_x.cost_matrix), axis=0)

  # Computes cost_matrix
  marginal_x = geom.marginal_from_potentials(f, g, axis=1)
  marginal_y = geom.marginal_from_potentials(f, g, axis=0)
  marginal_dep_term = _marginal_dependent_cost(
      marginal_x, marginal_y, geom_x, geom_y, loss)
  cost_matrix = marginal_dep_term - apply_cost_fn(geom_y)(
      tmp.T, axis=1, fn=loss.right_y).T
  return geometry.Geometry(
      cost_matrix=cost_matrix, epsilon=geom._epsilon_init, **kwargs)


def _marginal_dependent_cost(marginal_x, marginal_y, geom_x, geom_y, loss):
  r"""Calculates part of cost that depends on marginals of transport matrix.

  Uses the first term in Equation 6, Proposition 1 of
  http://proceedings.mlr.press/v48/peyre16.pdf.

  Let :math:`p` [num_a,] be the marginal of the transport matrix for samples
  from geom_x and :math:`q` [num_b,] be the marginal of the transport matrix for
  samples from geom_y. The term in the cost that depends on these marginals can
  be written as:
  marginal_dep_term = fn_x(cost_x) :math:`p \mathbb{1}_{num_b}^T`
                    + (fn_y(cost_y) :math:`q \mathbb{1}_{num_a}^T)^T`

  Args:
    marginal_x: jnp.ndarray<float>[num_a,], marginal of the transport matrix for
     samples from geom_x
    marginal_y: jnp.ndarray<float>[num_b,], marginal of the transport matrix for
     samples from geom_y
    geom_x: a Geometry object for the first view.
    geom_y: a second Geometry object for the second view.
    loss: a GWLossFn object.

  Returns:
    jnp.ndarray, [num_a, num_b]
  """
  x_term = jnp.dot(
      geom_x.apply_cost(marginal_x, 1, loss.fn_x).reshape(-1, 1),
      jnp.ones((1, marginal_y.size)))
  y_term = jnp.dot(
      geom_y.apply_cost(marginal_y, 1, loss.fn_y).reshape(-1, 1),
      jnp.ones((1, marginal_x.size))).T
  return x_term + y_term



# coding=utf-8
# Copyright 2021 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Low level functions used within the scope of Geometric processing."""

import functools
from typing import Sequence

import jax
import jax.numpy as jnp


@functools.partial(jax.custom_jvp, nondiff_argnums=(1, 2, 4))
def logsumexp(mat, axis=None, keepdims=False, b=None, return_sign=False):
  return jax.scipy.special.logsumexp(
      mat, axis=axis, keepdims=keepdims, b=b, return_sign=return_sign)


@logsumexp.defjvp
def logsumexp_jvp(axis, keepdims, return_sign, primals, tangents):
  """Implements custom derivative rule for lse that does not blow up with -inf.

  This logsumexp implementation uses the standard jax one in forward mode but
  implements a custom rule to differentiate. Given the preference of jax for
  jvp over vjp, and the fact that this is a simple linear rule, jvp is used.
  This custom differentiation address issues when the output of lse is
  -inf (which corresponds to the case where all inputs in a slice are -inf,
  which happens typically when ``a`` or ``b`` weight vectors have zeros.)

  Although both exp(lse) and its derivative should be 0, automatic
  differentiation returns a NaN derivative because of a -inf - (-inf) operation
  appearing in the definition of centered_exp below. This is corrected in the
  implementation below.

  Args:
    axis: argument from original logsumexp
    keepdims: argument from original logsumexp
    return_sign: argument from original logsumexp
    primals: mat and b, the two arguments against which we differentiate.
    tangents: of same size as mat and b.

  Returns:
    original primal outputs + their tangent.
  """
  mat, b = primals
  tan_mat, tan_b = tangents
  lse = logsumexp(mat, axis, keepdims, b, return_sign)
  if return_sign:
    lse, sign = lse
  lse = jnp.where(jnp.isfinite(lse), lse, 0.0)
  centered_exp = jnp.exp(mat - jnp.expand_dims(lse, axis=axis))

  if b is None:
    res = jnp.sum(centered_exp * tan_mat, axis=axis, keepdims=keepdims)
  else:
    res = jnp.sum(b * centered_exp * tan_mat, axis=axis, keepdims=keepdims)
    res += jnp.sum(tan_b * centered_exp, axis=axis, keepdims=keepdims)
  if return_sign:
    return (lse, sign), (sign * res, jnp.zeros_like(sign))
  else:
    return lse, res



# coding=utf-8
# Copyright 2021 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements the sinkhorn divergence."""

import collections
from typing import Any, Dict, Optional, Type

import jax
from jax import numpy as jnp
from ott.core import sinkhorn
from ott.geometry import geometry
from ott.geometry import pointcloud

SinkhornDivergenceOutput = collections.namedtuple(
    'SinkhornDivergenceOutput',
    ['divergence', 'potentials', 'geoms', 'errors', 'converged'])


def sinkhorn_divergence(
    geom: Type[geometry.Geometry],
    *args,
    a: Optional[jnp.ndarray] = None,
    b: Optional[jnp.ndarray] = None,
    sinkhorn_kwargs: Optional[Dict[str, Any]] = None,
    static_b: bool = False,
    share_epsilon: bool = True,
    **kwargs):
  """Computes Sinkhorn divergence defined by a geometry, weights, parameters.

  Args:
    geom: a geometry class.
    *args: arguments to the prepare_divergences method that is specific to each
      geometry.
    a: jnp.ndarray<float>[n]: the weight of each input point. The sum of
      all elements of b must match that of a to converge.
    b: jnp.ndarray<float>[m]: the weight of each target point. The sum of
      all elements of b must match that of a to converge.
    sinkhorn_kwargs: Optionally a dict containing the keywords arguments for
      calls to the `sinkhorn` function, that is called twice if static_b else
      three times.
    static_b: if True, divergence of measure b against itself is NOT computed
    share_epsilon: if True, enforces that the same epsilon regularizer is shared
      for all 2 or 3 terms of the Sinkhorn divergence. In that case, the epsilon
      will be by default that used when comparing x to y (contained in the first
      geometry). This flag is set to True by default, because in the default
      setting, the epsilon regularization is a function of the mean of the cost
      matrix.
    **kwargs: keywords arguments to the generic class. This is specific to each
      geometry.

  Returns:
    tuple: (sinkhorn divergence value, three pairs of potentials, three costs)
  """
  geometries = geom.prepare_divergences(*args, static_b=static_b, **kwargs)
  geometries = (geometries + (None,) * max(0, 3 - len(geometries)))[:3]
  if share_epsilon:
    for geom in filter(None, geometries[1:(2 if static_b else 3)]):
      geom.copy_epsilon(geometries[0])

  num_a, num_b = geometries[0].shape
  a = jnp.ones((num_a,)) / num_a if a is None else a
  b = jnp.ones((num_b,)) / num_b if b is None else b
  div_kwargs = {} if sinkhorn_kwargs is None else sinkhorn_kwargs
  return _sinkhorn_divergence(*geometries, a, b, **div_kwargs)


def _sinkhorn_divergence(
    geometry_xy: geometry.Geometry,
    geometry_xx: geometry.Geometry,
    geometry_yy: Optional[geometry.Geometry],
    a: jnp.ndarray,
    b: jnp.ndarray,
    **kwargs):
  """Computes the (unbalanced) sinkhorn divergence for the wrapper function.

    This definition includes a correction depending on the total masses of each
    measure, as defined in https://arxiv.org/pdf/1910.12958.pdf (15).

  Args:
    geometry_xy: a Cost object able to apply kernels with a certain epsilon,
    between the views X and Y.
    geometry_xx: a Cost object able to apply kernels with a certain epsilon,
    between elements of the view X.
    geometry_yy: a Cost object able to apply kernels with a certain epsilon,
    between elements of the view Y.
    a: jnp.ndarray<float>[n]: the weight of each input point. The sum of
     all elements of b must match that of a to converge.
    b: jnp.ndarray<float>[m]: the weight of each target point. The sum of
     all elements of b must match that of a to converge.
    **kwargs: Arguments to sinkhorn.
  Returns:
    SinkhornDivergenceOutput named tuple.
  """
  # When computing a Sinkhorn divergence, the (x,y) terms and (x,x) / (y,y)
  # terms are computed independently. The user might want to pass some
  # sinkhorn_kwargs to parameterize sinkhorn's behavior, but those should
  # only apply to the (x,y) part. For the (x,x) / (y,y) part we fall back
  # on a simpler choice (parallel_dual_updates + momentum 0.5) that is known
  # to work well in such settings. In the future we might want to give some
  # freedom on setting parameters for the (x,x)/(y,y) part.
  # Since symmetric terms are computed assuming a = b, the linear systems
  # arising in implicit differentiation (if used) of the potentials computed for
  # the symmetric parts should be marked as symmetric.
  kwargs_symmetric = kwargs.copy()
  kwargs_symmetric.update(
      parallel_dual_updates=True,
      momentum=0.5,
      chg_momentum_from=0,
      anderson_acceleration=0,
      implicit_solver_symmetric=True)

  out_xy = sinkhorn.sinkhorn(geometry_xy, a, b, **kwargs)
  out_xx = sinkhorn.sinkhorn(geometry_xx, a, a, **kwargs_symmetric)
  if geometry_yy is None:
    out_yy = sinkhorn.SinkhornOutput(None, None, 0, None, None)
  else:
    out_yy = sinkhorn.sinkhorn(geometry_yy, b, b, **kwargs_symmetric)

  div = (out_xy.reg_ot_cost - 0.5 * (out_xx.reg_ot_cost + out_yy.reg_ot_cost)
         + 0.5 * geometry_xy.epsilon * (jnp.sum(a) - jnp.sum(b))**2)
  out = (out_xy, out_xx, out_yy)
  return SinkhornDivergenceOutput(div, tuple([s.f, s.g] for s in out),
                                  (geometry_xy, geometry_xx, geometry_yy),
                                  tuple(s.errors for s in out),
                                  tuple(s.converged for s in out))


def segment_sinkhorn_divergence(x: jnp.ndarray,
                                y: jnp.ndarray,
                                segment_ids_x: Optional[jnp.ndarray] = None,
                                segment_ids_y: Optional[jnp.ndarray] = None,
                                num_segments: Optional[int] = None,
                                indices_are_sorted: Optional[bool] = None,
                                num_per_segment_x: Optional[jnp.ndarray] = None,
                                num_per_segment_y: Optional[jnp.ndarray] = None,
                                weights_x: Optional[jnp.ndarray] = None,
                                weights_y: Optional[jnp.ndarray] = None,
                                sinkhorn_kwargs: Optional[Dict[str,
                                                               Any]] = None,
                                static_b: bool = False,
                                share_epsilon: bool = True,
                                **kwargs) -> jnp.ndarray:
  """Computes Sinkhorn divergence between subsets of data with pointcloud.

  There are two interfaces: either use `segment_ids_x`, `segment_ids_y`, and
  optionally `num_segments` and `indices_are_sorted`, OR use `num_per_segment_x`
  and `num_per_segment_y`. If using the first interface, `num_segments` is
  required for JIT compilation. Assumes range(0, `num_segments`) are the segment
  ids. The second interface assumes `x` and `y` are segmented contiguously.

  In all cases, both `x` and `y` should contain the same number of segments.
  Each segment will be separately run through the sinkhorn divergence using
  array padding.

  Args:
    x: Array of input points, of shape [num_x, feature]. Multiple segments are
      held in this single array.
    y: Array of target points, of shape [num_y, feature].
    segment_ids_x: (1st interface) The segment ID for which each row of x
      belongs. This is a similar interface to `jax.ops.segment_sum`.
    segment_ids_y: (1st interface) The segment ID for which each row of y
      belongs.
    num_segments: (1st interface) Number of segments. This is required for JIT
      compilation to work. If not given, it will be computed from the data as
      the max segment ID.
    indices_are_sorted: (1st interface) Whether `segment_ids_x` and
      `segment_ids_y` are sorted. Default false.
    num_per_segment_x: (2nd interface) Number of points in each segment in `x`.
      For example, [100, 20, 30] would imply that `x` is segmented into three
      arrays of length `[100]`, `[20]`, and `[30]` respectively.
    num_per_segment_y: (2nd interface) Number of points in each segment in `y`.
    weights_x: Weights of each input points, arranged in the same segmented
      order as `x`.
    weights_y: Weights of each input points, arranged in the same segmented
      order as `y`.
    sinkhorn_kwargs: Optionally a dict containing the keywords arguments for
      calls to the `sinkhorn` function, that is called twice if static_b else
      three times.
    static_b: if True, divergence of measure b against itself is NOT computed
    share_epsilon: if True, enforces that the same epsilon regularizer is shared
      for all 2 or 3 terms of the Sinkhorn divergence. In that case, the epsilon
      will be by default that used when comparing x to y (contained in the first
      geometry). This flag is set to True by default, because in the default
      setting, the epsilon regularization is a function of the mean of the cost
      matrix.
    **kwargs: keywords arguments to the generic class. This is specific to each
      geometry.

  Returns:
    An array of sinkhorn divergence values for each segment.
  """

  use_segment_ids = segment_ids_x is not None

  if use_segment_ids:
    assert segment_ids_y is not None
  else:
    assert num_per_segment_x is not None
    assert num_per_segment_y is not None

  if use_segment_ids:
    if num_segments is None:
      num_segments = jnp.max(segment_ids_x) + 1
      assert num_segments == jnp.max(segment_ids_y) + 1

    if indices_are_sorted is None:
      indices_are_sorted = False

    num_per_segment_x = jax.ops.segment_sum(
        jnp.ones_like(segment_ids_x),
        segment_ids_x,
        num_segments=num_segments,
        indices_are_sorted=indices_are_sorted)
    num_per_segment_y = jax.ops.segment_sum(
        jnp.ones_like(segment_ids_y),
        segment_ids_y,
        num_segments=num_segments,
        indices_are_sorted=indices_are_sorted)
  else:
    assert num_segments is None
    num_segments = num_per_segment_x.shape[0]
    segment_ids_x = jnp.arange(num_segments).repeat(num_per_segment_x)
    segment_ids_y = jnp.arange(num_segments).repeat(num_per_segment_y)

  num_x = x.shape[0]
  num_y = y.shape[0]

  if weights_x is None:
    weights_x = (1 / num_per_segment_x).repeat(
        num_per_segment_x, total_repeat_length=num_x)

  if weights_y is None:
    weights_y = (1 / num_per_segment_y).repeat(
        num_per_segment_y, total_repeat_length=num_y)

  segmented_x = jnp.stack(
      [x * (segment_ids_x == i)[:, None] for i in range(num_segments)])

  segmented_y = jnp.stack(
      [y * (segment_ids_y == i)[:, None] for i in range(num_segments)])

  segmented_weights_x = jnp.stack(
      [weights_x * (segment_ids_x == i) for i in range(num_segments)])

  segmented_weights_y = jnp.stack(
      [weights_y * (segment_ids_y == i) for i in range(num_segments)])

  def single_segment_sink_div(padded_x, padded_y, padded_weight_x,
                              padded_weight_y):
    return sinkhorn_divergence(
        pointcloud.PointCloud,
        padded_x,
        padded_y,
        a=padded_weight_x,
        b=padded_weight_y,
        sinkhorn_kwargs=sinkhorn_kwargs,
        static_b=static_b,
        share_epsilon=share_epsilon,
        **kwargs).divergence

  v_sink_div = jax.vmap(single_segment_sink_div, in_axes=[0, 0, 0, 0])

  segmented_divergences = v_sink_div(segmented_x, segmented_y,
                                     segmented_weights_x, segmented_weights_y)
  return segmented_divergences



# coding=utf-8
# Copyright 2021 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A class describing operations used to instantiate and use a geometry."""
import functools
from typing import Optional, Sequence, Union

import jax
import jax.numpy as jnp
from ott.geometry import epsilon_scheduler
from ott.geometry import ops


@jax.tree_util.register_pytree_node_class
class Geometry:
  r"""Base class to define ground costs/kernels used in optimal transport.

  Optimal transport problems are intrinsically geometric: they compute an
  optimal way to transport mass from one configuration onto another. To define
  what is meant by optimality of a transport requires defining a cost, of moving
  mass from one among several sources, towards one out of multiple targets.
  These sources and targets can be provided as points in vectors spaces, grids,
  or more generally exclusively described through a (dissimilarity) cost matrix,
  or almost equivalently, a (similarity) kernel matrix.

  Once that cost or kernel matrix is set, the ``Geometry`` class provides a
  basic operations to be run with the Sinkhorn algorithm.

  Note:
    When defining a ``Geometry`` through a ``cost_matrix``, it is important to
    select an ``epsilon`` regularization parameter that is meaningful. That
    parameter can be provided by the user, or assigned a default value through
    a simple rule, using the mean cost value implied by the ``cost_matrix``.
  """

  def __init__(self,
               cost_matrix: Optional[jnp.ndarray] = None,
               kernel_matrix: Optional[jnp.ndarray] = None,
               epsilon: Union[epsilon_scheduler.Epsilon, float, None] = None,
               relative_epsilon: Optional[bool] = None,
               scale: float = None,
               **kwargs):
    r"""Initializes a geometry by passing it a cost matrix or a kernel matrix.

    Args:
      cost_matrix: jnp.ndarray<float>[num_a, num_b]: a cost matrix storing n x m
        costs.
      kernel_matrix: jnp.ndarray<float>[num_a, num_b]: a kernel matrix storing n
        x m kernel values.
      epsilon: a regularization parameter. If a ``epsilon_scheduler.Epsilon``
        object is passed, other parameters below are ignored in practice. If the
        parameter is a float, then this is understood to be the regularization
        that is needed, unless ``relative_epsilon`` below is ``True``, in which
        case ``epsilon`` is understood as a normalized quantity, to be scaled by
        the mean value of the ``cost_matrix``.
      relative_epsilon: whether epsilon is passed relative to scale of problem,
        here understood as mean value of ``cost_matrix``.
      scale: the scale multiplier for epsilon.
      **kwargs: additional kwargs to epsilon.
    """
    self._cost_matrix = cost_matrix
    self._kernel_matrix = kernel_matrix
    self._epsilon_init = epsilon
    self._relative_epsilon = relative_epsilon
    self._scale = scale
    # Define default dictionary and update it with user's values.
    self._kwargs = {**{'init': None, 'decay': None}, **kwargs}

  @property
  def scale(self) -> float:
    """Computes the scale of the epsilon, potentially based on data."""
    if isinstance(self._epsilon_init, epsilon_scheduler.Epsilon):
      return 1.0

    rel = self._relative_epsilon
    trigger = ((self._scale is None) and
               (rel or rel is None) and
               (self._epsilon_init is None or rel))
    if (self._scale is None) and (trigger is not None):  # for dry run
      return jnp.where(
          trigger, jax.lax.stop_gradient(self.mean_cost_matrix), 1.0)
    else:
      return self._scale

  @property
  def _epsilon(self):
    """Returns epsilon scheduler, either passed directly or by building it."""
    if isinstance(self._epsilon_init, epsilon_scheduler.Epsilon):
      return self._epsilon_init
    eps = 5e-2 if self._epsilon_init is None else self._epsilon_init
    return epsilon_scheduler.Epsilon.make(eps, scale=self.scale, **self._kwargs)

  @property
  def cost_matrix(self):
    """Returns cost matrix, computes it if only kernel was specified."""
    if self._cost_matrix is None:
      # If no epsilon was passed on to the geometry, then assume it is one by
      # default.
      cost = -jnp.log(self._kernel_matrix)
      return cost if self._epsilon_init is None else self.epsilon * cost
    return self._cost_matrix

  @property
  def median_cost_matrix(self):
    return jnp.median(self.cost_matrix)

  @property
  def mean_cost_matrix(self):
    if isinstance(self.shape[0], int) and (self.shape[0] > 0):
      return jnp.sum(self.apply_cost(jnp.ones((self.shape[0],)))) / (
          self.shape[0] * self.shape[1])
    else:
      return 1.0

  @property
  def kernel_matrix(self):
    if self._kernel_matrix is None:
      return jnp.exp(-(self._cost_matrix / self.epsilon))
    return self._kernel_matrix

  @property
  def epsilon(self):
    return self._epsilon.target

  @property
  def shape(self):
    mat = self.kernel_matrix if self.cost_matrix is None else self.cost_matrix
    if mat is not None:
      return mat.shape
    return (0, 0)

  @property
  def is_symmetric(self):
    mat = self.kernel_matrix if self.cost_matrix is None else self.cost_matrix
    return (mat.shape[0] == mat.shape[1] and
            jnp.all(mat == mat.T)) if mat is not None else False

  def copy_epsilon(self, other):
    """Copies the epsilon parameters from another geometry."""
    scheduler = other._epsilon
    self._epsilon_init = scheduler._target_init
    self._relative_epsilon = False
    self._scale = other.scale

  # The functions below are at the core of Sinkhorn iterations, they are
  # are implemented here in their default form, either in lse (using directly
  # cost matrices in stabilized form) or kernel mode (using kernel matrices).

  def apply_lse_kernel(self,
                       f: jnp.ndarray,
                       g: jnp.ndarray,
                       eps: float,
                       vec: jnp.ndarray = None,
                       axis: int = 0) -> jnp.ndarray:
    """Applies kernel in log domain on pair of dual potential variables.

    This function applies the ground geometry's kernel in log domain, using
    a stabilized formulation. At a high level, this iteration performs either

    output = eps * log (K (exp(g / eps) * vec) )    (1)
    or
    output = eps * log (K'(exp(f / eps) * vec))   (2)

    K is implicitly exp(-cost_matrix/eps).

    To carry this out in a stabilized way, we take advantage of the fact that
    the entries of the matrix f[:,*] + g[*,:] - C are all negative, and
    therefore their exponential never overflows, to add (and subtract after)
    f and g in iterations 1 & 2 respectively.

    Args:
      f: jnp.ndarray [num_a,] , potential of size num_rows of cost_matrix
      g: jnp.ndarray [num_b,] , potential of size num_cols of cost_matrix
      eps: float, regularization strength
      vec: jnp.ndarray [num_a or num_b,] , when not None, this has the effect of
        doing log-Kernel computations with an addition elementwise
        multiplication of exp(g /eps) by a vector. This is carried out by adding
        weights to the log-sum-exp function, and needs to handle signs
        separately.
      axis: summing over axis 0 when doing (2), or over axis 1 when doing (1)

    Returns:
      A jnp.ndarray corresponding to output above, depending on axis.
    """
    w_res, w_sgn = self._softmax(f, g, eps, vec, axis)
    remove = f if axis == 1 else g
    return w_res - jnp.where(jnp.isfinite(remove), remove, 0), w_sgn

  def apply_kernel(self, scaling: jnp.ndarray, eps=None, axis=0):
    """Applies kernel on positive scaling vector.

    This function applies the ground geometry's kernel, to perform either
    output = K v    (1)
    output = K'u   (2)
    where K is [num_a, num_b]

    Args:
      scaling: jnp.ndarray [num_a or num_b] , scaling of size num_rows or
        num_cols of kernel_matrix
      eps: passed for consistency, not used yet.
      axis: standard kernel product if axis is 1, transpose if 0.

    Returns:
      a jnp.ndarray corresponding to output above, depending on axis.
    """
    if eps is None:
      kernel = self.kernel_matrix
    else:
      kernel = self.kernel_matrix**(self.epsilon / eps)
    kernel = kernel if axis == 1 else kernel.T

    return jnp.dot(kernel, scaling)

  def marginal_from_potentials(self,
                               f: jnp.ndarray,
                               g: jnp.ndarray,
                               axis: int = 0) -> jnp.ndarray:
    """Outputs marginal of transportation matrix from potentials.

    This applies first lse kernel in the standard way, removes the
    correction used to stabilise computations, and lifts this with an exp to
    recover either of the marginals corresponding to the transport map induced
    by potentials.
    Args:
      f: jnp.ndarray [num_a,] , potential of size num_rows of cost_matrix
      g: jnp.ndarray [num_b,] , potential of size num_cols of cost_matrix
      axis: axis along which to integrate, returns marginal on other axis.

    Returns:
      a vector of marginals of the transport matrix.
    """
    h = (f if axis == 1 else g)
    z = self.apply_lse_kernel(f, g, self.epsilon, axis=axis)[0]
    return jnp.exp((z + h) / self.epsilon)

  def marginal_from_scalings(self, u: jnp.ndarray, v: jnp.ndarray, axis=0):
    """Outputs marginal of transportation matrix from scalings."""
    u, v = (v, u) if axis == 0 else (u, v)
    return u * self.apply_kernel(v, eps=self.epsilon, axis=axis)

  def transport_from_potentials(self, f, g):
    """Outputs transport matrix from potentials."""
    return jnp.exp(self._center(f, g) / self.epsilon)

  def transport_from_scalings(self, u, v):
    """Outputs transport matrix from pair of scalings."""
    return self.kernel_matrix * u[:, jnp.newaxis] * v[jnp.newaxis, :]

  # Functions that are not supposed to be changed by inherited classes.
  # These are the point of entry for Sinkhorn's algorithm to use a geometry.
  def error(self,
            f_u: jnp.ndarray,
            g_v: jnp.ndarray,
            target: jnp.ndarray,
            axis: int = 0,
            norm_error: Sequence[int] = (1,),
            lse_mode: bool = True):
    """Outputs error, using 2 potentials/scalings, of transport w.r.t target marginal.

    Args:
      f_u: a vector of potentials or scalings for the first marginal.
      g_v: a vector of potentials or scalings for the second marginal.
      target: target marginal.
      axis: axis (0 or 1) along which to compute marginal.
      norm_error: (t-uple of int) p's to compute p-norm between marginal/target
      lse_mode: whether operating on scalings or potentials

    Returns:
      t-uple of floats, quantifying difference between target / marginal.
    """
    if lse_mode:
      marginal = self.marginal_from_potentials(f_u, g_v, axis=axis)
    else:
      marginal = self.marginal_from_scalings(f_u, g_v, axis=axis)
    norm_error = jnp.array(norm_error)
    error = jnp.sum(
        jnp.abs(marginal - target)**norm_error[:, jnp.newaxis],
        axis=1)**(1.0 / norm_error)
    return error

  def update_potential(self, f, g, log_marginal, iteration=None, axis=0):
    """Carries out one Sinkhorn update for potentials, i.e.

    in log space.

    Args:
      f: jnp.ndarray [num_a,] , potential of size num_rows of cost_matrix
      g: jnp.ndarray [num_b,] , potential of size num_cols of cost_matrix
      log_marginal: targeted marginal
      iteration: used to compute epsilon from schedule, if provided.
      axis: axis along which the update should be carried out.

    Returns:
      new potential value, g if axis=0, f if axis is 1.
    """
    eps = self._epsilon.at(iteration)
    app_lse = self.apply_lse_kernel(f, g, eps, axis=axis)[0]
    return eps * log_marginal - jnp.where(jnp.isfinite(app_lse), app_lse, 0)

  def update_scaling(self, scaling, marginal, iteration=None, axis=0):
    """Carries out one Sinkhorn update for scalings, using kernel directly.

    Args:
      scaling: jnp.ndarray of num_a or num_b positive values.
      marginal: targeted marginal
      iteration: used to compute epsilon from schedule, if provided.
      axis: axis along which the update should be carried out.

    Returns:
      new scaling vector, of size num_b if axis=0, num_a if axis is 1.
    """

    eps = self._epsilon.at(iteration)
    app_kernel = self.apply_kernel(scaling, eps, axis=axis)
    return marginal / jnp.where(app_kernel > 0, app_kernel, 1.0)

  # Helper functions
  def _center(self, f: jnp.ndarray, g: jnp.ndarray):
    return f[:, jnp.newaxis] + g[jnp.newaxis, :] - self.cost_matrix

  def _softmax(self, f, g, eps, vec, axis):
    """Applies softmax row or column wise, weighted by vec."""
    if vec is not None:
      if axis == 0:
        vec = vec.reshape((vec.size, 1))
      lse_output = ops.logsumexp(
          self._center(f, g) / eps, b=vec, axis=axis, return_sign=True)
      return eps * lse_output[0], lse_output[1]
    else:
      lse_output = ops.logsumexp(
          self._center(f, g) / eps, axis=axis, return_sign=False)
      return eps * lse_output, jnp.array([1.0])

  @functools.partial(jax.vmap, in_axes=[None, None, None, 0, None])
  def _apply_transport_from_potentials(self, f, g, vec, axis):
    """Applies lse_kernel to arbitrary vector while keeping track of signs."""
    lse_res, lse_sgn = self.apply_lse_kernel(
        f, g, self.epsilon, vec=vec, axis=axis)
    lse_res += f if axis == 1 else g
    return lse_sgn * jnp.exp(lse_res / self.epsilon)

  # wrapper to allow default option for axis.
  def apply_transport_from_potentials(self,
                                      f: jnp.ndarray,
                                      g: jnp.ndarray,
                                      vec: jnp.ndarray,
                                      axis: int = 0) -> jnp.ndarray:
    """Applies transport matrix computed from potentials to a (batched) vec.

    This approach does not instantiate the transport matrix itself, but uses
    instead potentials to apply the transport using apply_lse_kernel, therefore
    guaranteeing stability and lower memory footprint.

    Computations are done in log space, and take advantage of the
    (b=..., return_sign=True) optional parameters of logsumexp.

    Args:
      f: jnp.ndarray [num_a,] , potential of size num_rows of cost_matrix
      g: jnp.ndarray [num_b,] , potential of size num_cols of cost_matrix
      vec: jnp.ndarray [batch, num_a or num_b], vector that will be multiplied
        by transport matrix corresponding to potentials f, g, and geom.
      axis: axis to differentiate left (0) or right (1) multiply.

    Returns:
      ndarray of the size of vec.
    """
    if vec.ndim == 1:
      return self._apply_transport_from_potentials(f, g, vec[jnp.newaxis, :],
                                                   axis)[0, :]
    return self._apply_transport_from_potentials(f, g, vec, axis)

  @functools.partial(jax.vmap, in_axes=[None, None, None, 0, None])
  def _apply_transport_from_scalings(self, u, v, vec, axis):
    u, v = (u, v * vec) if axis == 1 else (v, u * vec)
    return u * self.apply_kernel(v, eps=self.epsilon, axis=axis)

  # wrapper to allow default option for axis
  def apply_transport_from_scalings(self,
                                    u: jnp.ndarray,
                                    v: jnp.ndarray,
                                    vec: jnp.ndarray,
                                    axis: int = 0) -> jnp.ndarray:
    """Applies transport matrix computed from scalings to a (batched) vec.

    This approach does not instantiate the transport matrix itself, but
    relies instead on the apply_kernel function.

    Args:
      u: jnp.ndarray [num_a,] , scaling of size num_rows of cost_matrix
      v: jnp.ndarray [num_b,] , scaling of size num_cols of cost_matrix
      vec: jnp.ndarray [batch, num_a or num_b], vector that will be multiplied
        by transport matrix corresponding to scalings u, v, and geom.
      axis: axis to differentiate left (0) or right (1) multiply.

    Returns:
      ndarray of the size of vec.
    """
    if vec.ndim == 1:
      return self._apply_transport_from_scalings(u, v, vec[jnp.newaxis, :],
                                                 axis)[0, :]
    return self._apply_transport_from_scalings(u, v, vec, axis)

  def potential_from_scaling(self, scaling: jnp.ndarray) -> jnp.ndarray:
    return self.epsilon * jnp.log(scaling)

  def scaling_from_potential(self, potential: jnp.ndarray) -> jnp.ndarray:
    finite = jnp.isfinite(potential)
    return jnp.where(finite,
                     jnp.exp(jnp.where(finite, potential / self.epsilon, 0.0)),
                     0.0)

  def apply_cost(self, arr: jnp.ndarray, axis: int = 0, fn=None) -> jnp.ndarray:
    """Applies cost matrix to array (vector or matrix).

    This function applies the ground geometry's cost matrix, to perform either
    output = C arr (if axis=1)
    output = C' arr (if axis=0)
    where C is [num_a, num_b]

    Args:
      arr: jnp.ndarray [num_a or num_b, p], vector that will be multiplied by
        the cost matrix.
      axis: standard cost matrix if axis=1, transport if 0
      fn: function to apply to cost matrix element-wise before the dot product

    Returns:
      A jnp.ndarray, [num_b, p] if axis=0 or [num_a, p] if axis=1
    """
    if arr.ndim == 1:
      return jax.vmap(
          lambda x: self._apply_cost_to_vec(x, axis, fn),
          1,
          1,
      )(
          arr.reshape(-1, 1))
    return jax.vmap(
        lambda x: self._apply_cost_to_vec(x, axis, fn),
        1,
        1,
    )(
        arr)

  def _apply_cost_to_vec(self,
                         vec: jnp.ndarray,
                         axis: int = 0,
                         fn=None) -> jnp.ndarray:
    """Applies [num_a, num_b] fn(cost) (or transpose) to vector.

    Args:
      vec: jnp.ndarray [num_a,] ([num_b,] if axis=1) vector
      axis: axis on which the reduction is done.
      fn: function optionally applied to cost matrix element-wise, before the
        doc product

    Returns:
      A jnp.ndarray corresponding to cost x vector
    """
    matrix = self.cost_matrix.T if axis == 0 else self.cost_matrix
    matrix = fn(matrix) if fn is not None else matrix
    return jnp.dot(matrix, vec)

  @classmethod
  def prepare_divergences(cls, *args, static_b: bool = False, **kwargs):
    """Instantiates 2 (or 3) geometries to compute a Sinkhorn divergence."""
    size = 2 if static_b else 3
    nones = [None, None, None]
    cost_matrices = kwargs.pop('cost_matrix', args)
    kernel_matrices = kwargs.pop('kernel_matrix', nones)
    cost_matrices = cost_matrices if cost_matrices is not None else nones
    return tuple(
        cls(cost_matrix=arg1, kernel_matrix=arg2, **kwargs)
        for arg1, arg2, _ in zip(cost_matrices, kernel_matrices, range(size)))

  def tree_flatten(self):
    return (self._cost_matrix, self._kernel_matrix, self._epsilon_init,
            self._relative_epsilon, self._kwargs), None

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    del aux_data
    return cls(*children[:-1], **children[-1])



# coding=utf-8
# Copyright 2021 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A Jax version of Sinkhorn's algorithm."""

import collections
import functools
from typing import Any
from typing import Optional, Sequence, Tuple, Callable

import jax
import jax.numpy as jnp
import numpy as np
from ott.core import fixed_point_loop
from ott.geometry import geometry

SinkhornOutput = collections.namedtuple(
    'SinkhornOutput', ['f', 'g', 'reg_ot_cost', 'errors', 'converged'])


def sinkhorn(
    geom: geometry.Geometry,
    a: Optional[jnp.ndarray] = None,
    b: Optional[jnp.ndarray] = None,
    tau_a: float = 1.0,
    tau_b: float = 1.0,
    threshold: float = 1e-3,
    norm_error: int = 1,
    inner_iterations: int = 10,
    min_iterations: int = 0,
    max_iterations: int = 2000,
    momentum: float = 1.0,
    chg_momentum_from: int = 0,
    anderson_acceleration: int = 0,
    refresh_anderson_frequency: int = 1,
    lse_mode: bool = True,
    implicit_differentiation: bool = True,
    implicit_solver_fun=jax.scipy.sparse.linalg.cg,
    implicit_solver_ridge_kernel: float = 0.0,
    implicit_solver_ridge_identity: float = 0.0,
    implicit_solver_symmetric: bool = False,
    precondition_fun: Optional[Callable[[float], float]] = None,
    parallel_dual_updates: bool = False,
    use_danskin: bool = None,
    init_dual_a: Optional[jnp.ndarray] = None,
    init_dual_b: Optional[jnp.ndarray] = None,
    jit: bool = False) -> SinkhornOutput:
  r"""Solves regularized OT problem using Sinkhorn iterations.

  The Sinkhorn algorithm is a fixed point iteration that solves a regularized
  optimal transport (reg-OT) problem between two measures.
  The optimization variables are a pair of vectors (called potentials, or
  scalings when parameterized as exponentials of the former). Calling this
  function returns therefore a pair of optimal vectors. In addition to these,
  `sinkhorn` also returns the objective value achieved by these optimal vectors;
  a vector of size `max_iterations/inner_terations` that records the vector of
  values recorded to monitor convergence, throughout the execution of the
  algorithm (padded with ``-1`` if convergence happens before), as well as a
  boolean to signify whether the algorithm has converged within the number of
  iterations specified by the user.

  The reg-OT problem is specified by two measures, of respective sizes ``n`` and
  ``m``. From the viewpoint of the ``sinkhorn`` function, these two measures are
  only seen through a triplet (``geom``, ``a``, ``b``), where ``geom`` is a
  ``Geometry`` object, and ``a`` and ``b`` are weight vectors of respective
  sizes ``n`` and ``m``. Starting from two initial values for those potentials
  or scalings (both can be defined by the user by passing value in
  ``init_dual_a`` or ``init_dual_b``), the Sinkhorn algorithm will use
  elementary operations that are carried out by the ``geom`` object.

  Some maths:
    Given a geometry ``geom``, which provides a cost matrix :math:`C` with its
    regularization parameter :math:`\epsilon`, (resp. a kernel matrix :math:`K`)
    the reg-OT problem consists in finding two vectors `f`, `g` of size ``n``,
    ``m`` that maximize the following criterion.

    :math:`\arg\max_{f, g}{- <a, \phi_a^{*}(-f)> - <b, \phi_b^{*}(-g)> - \epsilon
    <e^{f/\epsilon}, e^{-C/\epsilon} e^{-g/\epsilon}}>`

    where :math:`\phi_a(z) = \rho_a z(\log z - 1)` is a scaled entropy, and
    :math:`\phi_a^{*}(z) = \rho_a e^{z/\varepsilon}` its Legendre transform.

    That problem can also be written, instead, using positive scaling vectors
    `u`, `v` of size ``n``, ``m``, handled with the kernel :math:`K:=e^{-C/\epsilon}`,
    :math:`\arg\max_{u, v >0} - <a,\phi_a^{*}(-\epsilon\log u)> + <b, \phi_b^{*}(-\epsilon\log v)> -  <u, K
    v>`

    Both of these problems corresponds, in their *primal* formulation, to solving the
    unbalanced optimal transport problem with a variable matrix `P` of size ``n``
    x ``m``:

    :math:`\arg\min_{P>0} <P,C> -\epsilon \text{KL}(P | ab^T) + \rho_a \text{KL}(P1 | a) + \rho_b \text{KL}(P^T1 | b)`

    where :math:`KL` is the generalized Kullback-Leibler divergence.

    The very same primal problem can also be written using a kernel :math:`K`
    instead of a cost :math:`C` as well:

    :math:`\arg\min_{P} \epsilon KL(P|K) + \rho_a \text{KL}(P1 | a) + \rho_b \text{KL}(P^T1 | b)`

    The *original* OT problem taught in linear programming courses is recovered
    by using the formulation above relying on the cost :math:`C`, and letting
    :math:`\epsilon \rightarrow 0`, and :math:`\rho_a, \rho_b \rightarrow \infty`.
    In that case the entropy disappears, whereas the :math:`KL` regularizations
    above become constraints on the marginals of :math:`P`: This results in a
    standard min cost flow problem. This problem is not handled for now in this
    toolbox, which focuses exclusively on the case :math:`\epsilon > 0`.

    The *balanced* regularized OT problem is recovered for finite
    :math:`\epsilon > 0` but letting :math:`\rho_a, \rho_b \rightarrow \infty`.
    This problem can be shown to be equivalent to a matrix scaling problem,
    which can be solved using the Sinkhorn fixed-point algorithm. To handle the
    case :math:`\rho_a, \rho_b \rightarrow \infty`, the ``sinkhorn`` function
    uses parameters ``tau_a`` := :math:`\rho_a / (\epsilon + \rho_a)` and
    ``tau_b`` := :math:`\rho_b / (\epsilon + \rho_b)` instead.
    Setting either of these parameters to 1 corresponds to setting the
    corresponding :math:`\rho_a, \rho_b` to :math:`\infty`.

    The Sinkhorn algorithm solves the reg-OT problem by seeking optimal `f`, `g`
    potentials (or alternatively their parameterization as positive scalings `u`,
    `v`), rather than solving the primal problem in :math:`P`. This is mostly for
    efficiency (potentials and scalings have a ``n + m`` memory footprint, rather
    than ``n m`` required to store `P`). This is also because both problems are,
    in fact, equivalent, since the optimal transport :math:`P^*` can be recovered
    from optimal potentials :math:`f^*`, :math:`g^*` or scalings :math:`u^*`,
    :math:`v^*`, using the geometry's cost or kernel matrix respectively:

      :math:`P^* = \exp\left(\frac{f^*\mathbf{1}_m^T + \mathbf{1}_n g^{*T} - C}{\epsilon}\right) \text{ or } P^* =
      \text{diag}(u^*) K \text{diag}(v^*)`

    By default, the Sinkhorn algorithm solves this dual problem in `f, g` or
    `u, v` using block coordinate ascent, i.e. devising an update for each `f`
    and `g` (resp. `u` and `v`) that cancels their respective gradients, one at
    a time. These two iterations are repeated ``inner_iterations`` times, after
    which the norm of these gradients will be evaluated and compared with the
    ``threshold`` value. The iterations are then repeated as long as that error
    exceeds ``threshold``.

  Note on Sinkhorn updates:
    The boolean flag ``lse_mode`` sets whether the algorithm is run in either:

      - log-sum-exp mode (``lse_mode=True``), in which case it is directly defined in terms of updates to `f` and `g`, using log-sum-exp computations. This requires access to the cost matrix :math:`C`, as it is stored, or possibly computed on the fly by ``geom``.

      - kernel mode (``lse_mode=False``), in which case it will require access to a matrix vector multiplication operator :math:`z \rightarrow K z`, where :math:`K` is either instantiated from :math:`C` as :math:`\exp(-C/\epsilon)`, or provided directly. In that case, rather than optimizing on :math:`f` and :math:`g`, it is more convenient to optimize on their so called scaling formulations, :math:`u := \exp(f / \epsilon)` and :math:`v := \exp(g / \epsilon)`. While faster (applying matrices is faster than applying ``lse`` repeatedly over lines), this mode is also less stable numerically, notably for smaller :math:`\epsilon`.

    In the source code, the variables ``f_u`` or ``g_v`` can be either regarded as potentials (real) or scalings (positive) vectors, depending on the choice of ``lse_mode`` by the user. Once optimization is carried out, we only return dual variables in potential form, i.e. ``f`` and ``g``.

    In addition to standard Sinkhorn updates, the user can also use heavy-ball type updates using a ``momentum`` parameter in ]0,2[. We also implement a strategy that tries to set that parameter adaptively ater ``chg_momentum_from`` iterations, as a function of progress in the error, as proposed in the literature.

    Another upgrade to the standard Sinkhorn updates provided to the users lies in using Anderson acceleration. This can be parameterized by setting the otherwise null ``anderson_acceleration`` to a positive integer. When selected,the algorithm will recompute, every ``refresh_anderson_frequency`` (set by default to 1) an extrapolation of the most recently computed ``anderson_acceleration`` iterates. When using that option, notice that differentiation (if required) can only be carried out using implicit differentiation, and that all momentum related parameters are ignored.

    The ``parallel_dual_updates`` flag is set to ``False`` by default. In that setting, ``g_v`` is first updated using the latest values for ``f_u`` and ``g_v``, before proceeding to update ``f_u`` using that new value for ``g_v``. When the flag is set to ``True``, both ``f_u`` and ``g_v`` are updated simultaneously. Note that setting that choice to ``True`` requires using some form of averaging (e.g. ``momentum=0.5``). Without this, and on its own ``parallel_dual_updates`` won't work.

  Differentiation:
    The optimal solutions ``f`` and ``g`` and the optimal objective (``reg_ot_cost``) outputted by the Sinkhorn algorithm can be differentiated w.r.t. relevant inputs ``geom``, ``a`` and ``b`` using, by default, implicit differentiation of the optimality conditions (``implicit_differentiation`` set to ``True``). This choice has two consequences.

      - The termination criterion used to stop Sinkhorn (cancellation of gradient of objective w.r.t. ``f_u`` and ``g_v``) is used to differentiate ``f`` and ``g``, given a change in the inputs. These changes are computed by solving a linear system. The arguments starting with ``implicit_solver_*`` allow to define the linear solver that is used, and to control for two types or regularization (we have observed that, depending on the architecture, linear solves may require higher ridge parameters to remain stable). The optimality conditions in Sinkhorn can be analyzed as satisfying a ``z=z'`` condition, which are then differentiated. It might be beneficial (e.g. as in https://arxiv.org/abs/2002.03229) to use a preconditionning function ``precondition_fun`` to differentiate instead ``h(z)=h(z')``.

      - The objective ``reg_ot_cost`` returned by Sinkhon uses the so-called enveloppe (or Danskin's) theorem. In that case, because it is assumed that the gradients of the dual variables ``f_u`` and ``g_v`` w.r.t. dual objective are zero (reflecting the fact that they are optimal), small variations in ``f_u`` and ``g_v`` due to changes in inputs (such as ``geom``, ``a`` and ``b``) are considered negligible. As a result, ``stop_gradient`` is applied on dual variables ``f_u`` and ``g_v`` when evaluating the ``reg_ot_cost`` objective. Note that this approach is `invalid` when computing higher order derivatives. In that case the ``use_danskin`` flag must be set to ``False``.

    An alternative yet more costly way to differentiate the outputs of the Sinkhorn iterations is to use unrolling, i.e. reverse mode differentiation of the Sinkhorn loop. This is possible because Sinkhorn iterations are wrapped in a custom fixed point iteration loop, defined in ``fixed_point_loop``, rather than a standard while loop. This is to ensure the end result of this fixed point loop can also be differentiated, if needed, using standard JAX operations. To ensure backprop differentiability, the ``fixed_point_loop.fixpoint_iter_backprop`` loop does checkpointing of state variables (here ``f_u`` and ``g_v``) every ``inner_iterations``, and backpropagates automatically, block by block, through blocks of ``inner_iterations`` at a time.

  Note:
    * The Sinkhorn algorithm may not converge within the maximum number of iterations for possibly several reasons:

      1. the regularizer (defined as ``epsilon`` in the geometry ``geom`` object) is too small. Consider either switching to ``lse_mode=True`` (at the price of a slower execution), increasing ``epsilon``, or, alternatively, if you are unable or unwilling to increase  ``epsilon``, either increase ``max_iterations`` or ``threshold``.
      2. the probability weights ``a`` and ``b`` do not have the same total mass, while using a balanced (``tau_a=tau_b=1.0``) setup. Consider either normalizing ``a`` and ``b``, or set either ``tau_a`` and/or ``tau_b<1.0``.
      3. OOMs issues may arise when storing either cost or kernel matrices that are too large in ``geom``. In the case where, the ``geom`` geometry is a ``PointCloud``, some of these issues might be solved by setting the ``online`` flag to ``True``. This will trigger a recomputation on the fly of the cost/kernel matrix.

    * The weight vectors ``a`` and ``b`` can be passed on with coordinates that have zero weight. This is then handled by relying on simple arithmetic for ``inf`` values that will likely arise (due to :math:`log(0)` when ``lse_mode`` is ``True``, or divisions by zero when ``lse_mode`` is ``False``). Whenever that arithmetic is likely to produce ``NaN`` values (due to ``-inf * 0``, or ``-inf - -inf``) in the forward pass, we use ``jnp.where`` conditional statements to carry ``inf`` rather than ``NaN`` values. In the reverse mode differentiation, the inputs corresponding to these 0 weights (a location `x`, or a row in the corresponding cost/kernel matrix), and the weight itself will have ``NaN`` gradient values. This is reflects that these gradients are undefined, since these points were not considered in the optimization and have therefore no impact on the output.

  Args:
    geom: a Geometry object.
    a: [num_a,] or [batch, num_a] weights.
    b: [num_b,] or [batch, num_b] weights.
    tau_a: ratio rho/(rho+eps) between KL divergence regularizer to first
     marginal and itself + epsilon regularizer used in the unbalanced
     formulation.
    tau_b: ratio rho/(rho+eps) between KL divergence regularizer to first
     marginal and itself + epsilon regularizer used in the unbalanced
     formulation.
    threshold: tolerance used to stop the Sinkhorn iterations. This is
     typically the deviation between a target marginal and the marginal of the
     current primal solution when either or both tau_a and tau_b are 1.0
     (balanced or semi-balanced problem), or the relative change between two
     successive solutions in the unbalanced case.
    norm_error: power used to define p-norm of error for marginal/target.
    inner_iterations: the Sinkhorn error is not recomputed at each
     iteration but every inner_num_iter instead.
    min_iterations: the minimum number of Sinkhorn iterations carried
     out before the error is computed and monitored.
    max_iterations: the maximum number of Sinkhorn iterations. If
      ``max_iterations`` is equal to ``min_iterations``, sinkhorn iterations are
      run by default using a ``jax.lax.scan`` loop rather than a custom,
      unroll-able ``jax.lax.while_loop`` that monitors convergence. In that case
      the error is not monitored and the ``converged`` flag will return
      ``False`` as a consequence.
    momentum: a float in ]0,2[
    chg_momentum_from: if positive, momentum is recomputed using the
      adaptive rule provided in https://arxiv.org/pdf/2012.12562v1.pdf after
      that number of iterations.
    anderson_acceleration: int, if 0 (default), no acceleration. If positive,
      use Anderson acceleration on the dual sinkhorn (in log/potential form), as
      described in https://en.wikipedia.org/wiki/Anderson_acceleration and
      advocated in https://arxiv.org/abs/2006.08172, with a memory of size equal
      to ``anderson_acceleration``. In that case, differentiation is
      necessarly handled implicitly (``implicit_differentiation`` is set to
      ``True``) and all ``momentum`` related parameters are ignored.
    refresh_anderson_frequency: int, when using ``anderson_acceleration``,
      recompute direction periodically every int sinkhorn iterations.
    lse_mode: ``True`` for log-sum-exp computations, ``False`` for kernel
      multiplication.
    implicit_differentiation: ``True`` if using implicit differentiation,
      ``False`` if unrolling Sinkhorn iterations.
    implicit_solver_fun: see ``solve_implicit_system``
    implicit_solver_ridge_kernel: see ``solve_implicit_system``
    implicit_solver_ridge_identity: see ``solve_implicit_system``
    implicit_solver_symmetric: see ``solve_implicit_system``
    precondition_fun: Callable to modify FOC before differentiating them in
      implicit differentiation.
    parallel_dual_updates: updates potentials or scalings in parallel if True,
      sequentially (in Gauss-Seidel fashion) if False.
    use_danskin: when ``True``, it is assumed the entropy regularized cost is
      is evaluated using optimal potentials that are freezed, i.e. whose
      gradients have been stopped. This is useful when carrying out first order
      differentiation, and is only valid (as with ``implicit_differentiation``)
      when the algorithm has converged with a low tolerance.
    init_dual_a: optional initialization for potentials/scalings w.r.t.
      first marginal (``a``) of reg-OT problem.
    init_dual_b: optional initialization for potentials/scalings w.r.t.
      second marginal (``b``) of reg-OT problem.
    jit: if True, automatically jits the function upon first call.
      Should be set to False when used in a function that is jitted by the user,
      or when computing gradients (in which case the gradient function
      should be jitted by the user)

  Returns:
    a ``SinkhornOutput`` named tuple. The tuple contains two optimal potential
    vectors ``f`` and ``g``, the objective ``reg_ot_cost`` evaluated at those
    solutions, an array of ``errors`` to monitor convergence every
    ``inner_iterations`` and a flag ``converged`` that is ``True`` if the
    algorithm has converged within the number of iterations that was predefined
    by the user.
  """
  # Start by checking inputs.
  num_a, num_b = geom.shape
  a = jnp.ones((num_a,)) / num_a if a is None else a
  b = jnp.ones((num_b,)) / num_b if b is None else b

  if init_dual_a is None:
    init_dual_a = jnp.zeros_like(a) if lse_mode else jnp.ones_like(a)

  if init_dual_b is None:
    init_dual_b = jnp.zeros_like(b) if lse_mode else jnp.ones_like(b)

  if precondition_fun is None:
    precondition_fun = lambda x: geom.epsilon * jnp.log(x)

  # Cancel dual variables for zero weights.
  init_dual_a = jnp.where(a > 0, init_dual_a, -jnp.inf if lse_mode else 0.0)
  init_dual_b = jnp.where(b > 0, init_dual_b, -jnp.inf if lse_mode else 0.0)

  # Force implicit_differentiation to True when using Anderson acceleration,
  # Reset all momentum parameters.
  if anderson_acceleration:
    implicit_differentiation = True
    momentum = 1.0
    chg_momentum_from = 0

  # To change momentum adaptively, one needs errors in ||.||_1 norm.
  # In that case, we add this exponent to the list of errors to compute,
  # notably if that was not the error requested by the user.
  if chg_momentum_from > 0 and norm_error != 1:
    norm_error = (norm_error, 1)
  else:
    norm_error = (norm_error,)

  if jit:
    call_to_sinkhorn = functools.partial(
        jax.jit, static_argnums=(3, 4, 6, 7, 8, 9) + tuple(range(11, 23)))(
            _sinkhorn)
  else:
    call_to_sinkhorn = _sinkhorn
  return call_to_sinkhorn(geom, a, b, tau_a, tau_b, threshold, norm_error,
                          inner_iterations, min_iterations, max_iterations,
                          momentum, chg_momentum_from, anderson_acceleration,
                          refresh_anderson_frequency,
                          lse_mode, implicit_differentiation,
                          implicit_solver_fun,
                          implicit_solver_ridge_kernel,
                          implicit_solver_ridge_identity,
                          implicit_solver_symmetric,
                          precondition_fun,
                          parallel_dual_updates,
                          use_danskin, init_dual_a, init_dual_b)


def _sinkhorn(
    geom: geometry.Geometry,
    a: jnp.ndarray,
    b: jnp.ndarray,
    tau_a: float,
    tau_b: float,
    threshold: float,
    norm_error: int,
    inner_iterations: int,
    min_iterations: int,
    max_iterations: int,
    momentum: float,
    chg_momentum_from: int,
    anderson_acceleration: int,
    refresh_anderson_frequency: int,
    lse_mode: bool,
    implicit_differentiation: bool,
    implicit_solver_fun: Callable,
    implicit_solver_ridge_kernel: float,
    implicit_solver_ridge_identity: float,
    implicit_solver_symmetric: bool,
    precondition_fun: Callable[[float], float],
    parallel_dual_updates: bool,
    use_danskin: bool,
    init_dual_a: jnp.ndarray,
    init_dual_b: jnp.ndarray) -> SinkhornOutput:
  """Forks between implicit/backprop exec of Sinkhorn."""

  if implicit_differentiation:
    iteration_fun = _sinkhorn_iterations_implicit
  else:
    iteration_fun = _sinkhorn_iterations

  # By default, use Danskin theorem to differentiate
  # the objective when using implicit_differentiation.
  use_danskin = implicit_differentiation if use_danskin is None else use_danskin

  f, g, errors = iteration_fun(
      tau_a, tau_b, inner_iterations, min_iterations, max_iterations,
      chg_momentum_from, anderson_acceleration, refresh_anderson_frequency,
      lse_mode, implicit_differentiation, implicit_solver_fun,
      implicit_solver_ridge_kernel, implicit_solver_ridge_identity,
      implicit_solver_symmetric, precondition_fun,
      parallel_dual_updates, init_dual_a, init_dual_b, momentum, threshold,
      norm_error, geom, a, b)

  # When differentiating the regularized OT cost, and assuming Sinkhorn has run
  # to convergence, Danskin's (or the enveloppe) theorem
  # https://en.wikipedia.org/wiki/Danskin%27s_theorem
  # states that the resulting OT cost as a function of any of the inputs
  # (``geometry``, ``a``, ``b``) behaves locally as if the dual optimal
  # potentials were frozen and did not vary with those inputs.
  #
  # Notice this is only valid, as when using ``implicit_differentiation`` mode,
  # if the Sinkhorn algorithm outputs potentials that are near optimal.
  # namely when the threshold value is set to a small tolerance.
  #
  # The flag ``use_danskin`` controls whether that assumption is made. By
  # default, that flag is set to the value of ``implicit_differentiation`` if
  # not specified. If you wish to compute derivatives of order 2 and above,
  # set ``use_danskin`` to ``False``.
  reg_ot_cost = ent_reg_cost(geom, a, b, tau_a, tau_b,
                             jax.lax.stop_gradient(f) if use_danskin else f,
                             jax.lax.stop_gradient(g) if use_danskin else g,
                             lse_mode)
  converged = jnp.logical_and(
      jnp.sum(errors == -1) > 0,
      jnp.sum(jnp.isnan(errors)) == 0)

  return SinkhornOutput(f, g, reg_ot_cost, errors, converged)


def _sinkhorn_iterations(
    tau_a: float,
    tau_b: float,
    inner_iterations: int,
    min_iterations: int,
    max_iterations: int,
    chg_momentum_from: int,
    anderson_acceleration: int,
    refresh_anderson_frequency: int,
    lse_mode: bool,
    implicit_differentiation: bool,
    implicit_solver_fun: Callable,
    implicit_solver_ridge_kernel: float,
    implicit_solver_ridge_identity: float,
    implicit_solver_symmetric: bool,
    precondition_fun: Callable[[float], float],
    parallel_dual_updates: bool,
    init_dual_a: jnp.ndarray,
    init_dual_b: jnp.ndarray,
    momentum: float,
    threshold: float,
    norm_error: Sequence[int],
    geom: geometry.Geometry,
    a: jnp.ndarray,
    b: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
  r"""A jit'able Sinkhorn loop.

  For a detailed explanation of args, see parent function ``sinkhorn``
  Args:
    tau_a: float, ratio lam/(lam+eps) between KL divergence regularizer to first
      marginal and itself + epsilon regularizer used in the unbalanced
      formulation.
    tau_b: float, ratio lam/(lam+eps) between KL divergence regularizer to first
      marginal and itself + epsilon regularizer used in the unbalanced
      formulation.
    inner_iterations: (int32) the Sinkhorn error is not recomputed at each
      iteration but every inner_num_iter instead.
    min_iterations: (int32) the minimum number of Sinkhorn iterations.
    max_iterations: (int32) the maximum number of Sinkhorn iterations.
    chg_momentum_from: int, # of iterations after which momentum is computed
    anderson_acceleration: int, memory size for anderson acceleration.
    refresh_anderson_frequency: int, used every iterations to refresh.
    lse_mode: True for log-sum-exp computations, False for kernel
      multiplication.
    implicit_differentiation: if True, do not backprop through the Sinkhorn
      loop, but use the implicit function theorem on the fixed point optimality
      conditions.
    implicit_solver_fun: see ``solve_implicit_system``
    implicit_solver_ridge_kernel: see ``solve_implicit_system``
    implicit_solver_ridge_identity: see ``solve_implicit_system``
    implicit_solver_symmetric: see ``solve_implicit_system``
    precondition_fun: preconditioning function to stabilize FOC system.
    parallel_dual_updates: updates potentials or scalings in parallel if True,
      sequentially (in Gauss-Seidel fashion) if False.
    init_dual_a: optional initialization for potentials/scalings w.r.t. first
      marginal (``a``) of reg-OT problem.
    init_dual_b: optional initialization for potentials/scalings w.r.t. second
      marginal (``b``) of reg-OT problem.
    momentum: float, a float between ]0,2[
    threshold: (float) the relative threshold on the Sinkhorn error to stop the
      Sinkhorn iterations.
    norm_error: t-uple of int, p-norms of marginal / target errors to track
    geom: a Geometry object.
    a: jnp.ndarray<float>[num_a,] or jnp.ndarray<float>[batch,num_a] weights.
    b: jnp.ndarray<float>[num_b,] or jnp.ndarray<float>[batch,num_b] weights.

  Returns:
    f: potential
    g: potential
    errors: ndarray of errors
  """
  # Initializing solutions
  f_u, g_v = init_dual_a, init_dual_b

  # Delete arguments not used in forward pass.
  del implicit_solver_fun, implicit_solver_ridge_kernel
  del implicit_solver_ridge_identity, implicit_solver_symmetric
  del precondition_fun

  # Defining the Sinkhorn loop, by setting initializations, body/cond.
  errors = -jnp.ones(
      (np.ceil(max_iterations / inner_iterations).astype(int), len(norm_error)))
  const = (geom, a, b, threshold)

  def cond_fn(iteration, const, state):
    threshold = const[-1]
    errors = state[0]
    err = errors[iteration // inner_iterations - 1, 0]

    return jnp.logical_or(iteration == 0,
                          jnp.logical_and(jnp.isfinite(err), err > threshold))

  def get_momentum(errors, idx):
    """momentum formula, https://arxiv.org/pdf/2012.12562v1.pdf, p.7 and (5)."""
    error_ratio = jnp.minimum(errors[idx - 1, -1] / errors[idx - 2, -1], .99)
    power = 1.0 / inner_iterations
    return 2.0 / (1.0 + jnp.sqrt(1.0 - error_ratio**power))

  def anderson_extrapolation(xs, fxs, ridge_identity=1e-2):
    """Computes Anderson extrapolation from past observations."""
    # Remove -inf values to instantiate quadratic problem. All others
    # remain since they might be caused by a valid issue.
    fxs_clean = jnp.nan_to_num(fxs, nan=jnp.nan, posinf=jnp.inf, neginf=0.0)
    xs_clean = jnp.nan_to_num(xs, nan=jnp.nan, posinf=jnp.inf, neginf=0.0)
    residuals = fxs_clean - xs_clean
    gram_matrix = jnp.matmul(residuals.T, residuals)
    gram_matrix /= jnp.linalg.norm(gram_matrix)

    # Solve linear system to obtain weights
    weights = jax.scipy.sparse.linalg.cg(
        gram_matrix + ridge_identity * jnp.eye(xs.shape[1]),
        jnp.ones(xs.shape[1]))[0]
    weights /= jnp.sum(weights)

    # Recover linear combination and return it with NaN (caused
    # by 0 weights leading to -jnp.inf potentials, mixed with weights
    # coefficiences of different signs), disambiguated to -inf.
    combination = jnp.sum(fxs * weights[None, :], axis=1)
    return jnp.where(jnp.isfinite(combination), combination, -jnp.inf)

  def anderson_update(iteration, f_u, old_f_u_s, old_mapped_f_u_s):
    # Anderson acceleration always happens in potentials (not scalings) space,
    # regardless of the lse_mode setting. If the iteration count is large
    # enough, and update_anderson is True, the update below will output a
    # potential variable.

    trigger_update = jnp.logical_and(
        iteration > anderson_acceleration,
        iteration % refresh_anderson_frequency == 0)

    f_u = jnp.where(
        trigger_update,
        anderson_extrapolation(old_f_u_s, old_mapped_f_u_s),
        f_u)

    # If the interpolation was triggered, we store it in memory
    # Otherwise we add the previous value (converting it to potential form if
    # it was initially stored in scaling form).

    old_f_u_s = jnp.where(
        trigger_update,
        jnp.concatenate((old_f_u_s[:, 1:], f_u[:, None]), axis=1),
        jnp.concatenate(
            (old_f_u_s[:, 1:],
             (f_u if lse_mode
              else geom.potential_from_scaling(f_u))[:, None]),
            axis=1))

    # If update was triggered, ensure a scaling is returned, since the result
    # from the extrapolation was outputted in potential form.
    f_u = jnp.where(
        trigger_update,
        f_u if lse_mode else geom.scaling_from_potential(f_u),
        f_u)
    return f_u, old_f_u_s

  def body_fn(iteration, const, state, compute_error):
    """Carries out sinkhorn iteration.

    Depending on lse_mode, these iterations can be either in:
      - log-space for numerical stability.
      - scaling space, using standard kernel-vector multiply operations.

    Args:
      iteration: iteration number
      const: tuple of constant parameters that do not change throughout the
        loop, here the geometry and the marginals a, b.
      state: error log, potential/scaling variables updated in the loop, and
        history for Anderson acceleration if selected.
      compute_error: flag to indicate this iteration computes/stores an error

    Returns:
      state variables, i.e. errors and updated f_u, g_v potentials + history.
    """
    geom, a, b, _ = const
    errors, f_u, g_v, old_f_u_s, old_mapped_f_u_s = state

    # compute momentum term if needed, using previously seen errors.
    w = jax.lax.stop_gradient(
        jnp.where(
            iteration >= jnp.where(chg_momentum_from == 0, jnp.inf,
                                   chg_momentum_from),
            get_momentum(errors, chg_momentum_from // inner_iterations),
            momentum))

    # Sinkhorn updates using momentum or Anderson acceleration,
    # in either scaling or potential form.

    # When running updates in parallel (Gauss-Seidel mode), old_g_v will be
    # used to update f_u, rather than the latest g_v computed in this loop.
    if parallel_dual_updates:
      old_g_v = g_v

    # When using Anderson acceleration, first update the dual variable f_u with
    # previous updates (if iteration count sufficiently large), then record
    # new iterations in array.
    if anderson_acceleration:
      f_u, old_f_u_s = anderson_update(
          iteration, f_u, old_f_u_s, old_mapped_f_u_s)

    # In lse_mode, run additive updates.
    if lse_mode:
      new_g_v = tau_b * geom.update_potential(
          f_u, g_v, jnp.log(b), iteration, axis=0)
      g_v = (1.0 - w) * jnp.where(jnp.isfinite(g_v), g_v, 0.0) + w * new_g_v
      new_f_u = tau_a * geom.update_potential(
          f_u,
          old_g_v if parallel_dual_updates else g_v,
          jnp.log(a),
          iteration,
          axis=1)
      f_u = (1.0 - w) * jnp.where(jnp.isfinite(f_u), f_u, 0.0) + w * new_f_u
    # In kernel mode, run multiplicative updates.
    else:
      new_g_v = geom.update_scaling(f_u, b, iteration, axis=0)**tau_b
      g_v = jnp.where(g_v > 0, g_v, 1)**(1.0 - w) * new_g_v**w
      new_f_u = geom.update_scaling(
          old_g_v if parallel_dual_updates else g_v, a, iteration,
          axis=1)**tau_a
      f_u = jnp.where(f_u > 0, f_u, 1)**(1.0 - w) * new_f_u**w

    # When using Anderson acceleration, refresh latest update.
    if anderson_acceleration:
      old_mapped_f_u_s = jnp.concatenate(
          (old_mapped_f_u_s[:, 1:],
           (f_u if lse_mode else geom.potential_from_scaling(f_u))[:, None]),
          axis=1)

    # re-computes error if compute_error is True, else set it to inf.
    err = jnp.where(
        jnp.logical_and(compute_error, iteration >= min_iterations),
        marginal_error(geom, a, b, tau_a, tau_b, f_u, g_v, norm_error,
                       lse_mode), jnp.inf)

    errors = errors.at[iteration // inner_iterations, :].set(err)
    return errors, f_u, g_v, old_f_u_s, old_mapped_f_u_s

  # Run the Sinkhorn loop. choose either a standard fixpoint_iter loop if
  # differentiation is implicit, otherwise switch to the backprop friendly
  # version of that loop if unrolling to differentiate.

  if implicit_differentiation:
    fix_point = fixed_point_loop.fixpoint_iter
  else:
    fix_point = fixed_point_loop.fixpoint_iter_backprop

  # Initialize log matrix used in Anderson acceleration with nan values.
  # these values will be replaced by actual iteration values.
  if anderson_acceleration:
    old_f_u_s = jnp.ones(
        (geom.shape[0], anderson_acceleration)) * jnp.nan
    old_mapped_f_u_s = old_f_u_s
  else:
    old_f_u_s, old_mapped_f_u_s = None, None

  errors, f_u, g_v, _, _ = fix_point(
      cond_fn, body_fn, min_iterations, max_iterations, inner_iterations, const,
      (errors, f_u, g_v, old_f_u_s, old_mapped_f_u_s))

  f = f_u if lse_mode else geom.potential_from_scaling(f_u)
  g = g_v if lse_mode else geom.potential_from_scaling(g_v)

  return f, g, errors[:, 0]


def _sinkhorn_iterations_taped(
    tau_a: float,
    tau_b: float,
    inner_iterations: int,
    min_iterations: int,
    max_iterations: int,
    chg_momentum_from: int,
    anderson_acceleration: int,
    refresh_anderson_frequency: int,
    lse_mode: bool,
    implicit_differentiation: bool,
    implicit_solver_fun: Callable,
    implicit_solver_ridge_kernel: float,
    implicit_solver_ridge_identity: float,
    implicit_solver_symmetric: bool,
    precondition_fun: Optional[Callable[[float], float]],
    parallel_dual_updates: bool,
    init_dual_a: jnp.ndarray,
    init_dual_b: jnp.ndarray,
    momentum: float,
    threshold: float,
    norm_error: Sequence[int],
    geom: geometry.Geometry,
    a: jnp.ndarray,
    b: jnp.ndarray):
  """Runs forward pass of the Sinkhorn algorithm storing side information."""
  f, g, errors = _sinkhorn_iterations(
      tau_a, tau_b, inner_iterations, min_iterations, max_iterations,
      chg_momentum_from, anderson_acceleration, refresh_anderson_frequency,
      lse_mode, implicit_differentiation,
      implicit_solver_fun,
      implicit_solver_ridge_kernel,
      implicit_solver_ridge_identity,
      implicit_solver_symmetric,
      precondition_fun,
      parallel_dual_updates, init_dual_a, init_dual_b, momentum,
      threshold, norm_error, geom, a, b)
  return (f, g, errors), (f, g, geom, a, b)


def _sinkhorn_iterations_implicit_bwd(
    tau_a, tau_b, inner_iterations, min_iterations, max_iterations,
    chg_momentum_from, anderson_acceleration, refresh_anderson_frequency,
    lse_mode, implicit_differentiation, implicit_solver_fun,
    implicit_solver_ridge_kernel, implicit_solver_ridge_identity,
    implicit_solver_symmetric, precondition_fun,
    parallel_dual_updates, res, gr
) -> Tuple[Any, Any, Any, Any, geometry.Geometry, jnp.ndarray, jnp.ndarray]:
  """Runs Sinkhorn in backward mode, using implicit differentiation.

  Args:
    tau_a: float
    tau_b: float.
    inner_iterations: iters
    min_iterations: minimum number of Sinkhorn iterations.
    max_iterations: maximum number of Sinkhorn iterations.
    chg_momentum_from: int
    anderson_acceleration: int
    refresh_anderson_frequency: int
    lse_mode: True for log-sum-exp computations, False for kernel
      multiplication.
    implicit_differentiation: implicit or backprop.
    implicit_solver_fun: see ``solve_implicit_system``
    implicit_solver_ridge_kernel: see ``solve_implicit_system``
    implicit_solver_ridge_identity: see ``solve_implicit_system``
    implicit_solver_symmetric: see ``solve_implicit_system``
    precondition_fun: function to apply to the two blocks of the first
      order condition.
    parallel_dual_updates: update sequence.
    res: residual data sent from fwd pass, used for computations below. In this
      case consists in the output itself, as well as inputs against which we
      wish to differentiate.
    gr: gradients w.r.t outputs of fwd pass, here w.r.t size f, g, errors. Note
      that differentiability w.r.t. errors is not handled, and only f, g is
      considered.

  Returns:
    a tuple of gradients: PyTree for geom, one jnp.ndarray for each of a and b.
  """
  del inner_iterations, min_iterations, max_iterations
  del chg_momentum_from, anderson_acceleration, refresh_anderson_frequency
  del implicit_differentiation, parallel_dual_updates

  f, g, geom, a, b = res
  # Ignores gradients info with respect to 'errors' output.
  gr = gr[0], gr[1]

  # Applies first part of vjp to gr: inverse part of implicit function theorem.
  vjp_gr = solve_implicit_system(gr, geom, a, b, f, g, tau_a, tau_b, lse_mode,
                                 precondition_fun,
                                 implicit_solver_fun,
                                 implicit_solver_ridge_kernel,
                                 implicit_solver_ridge_identity,
                                 implicit_solver_symmetric)

  # Instantiates vjp of first order conditions of the objective, as a
  # function of geom, a and b parameters (against which we differentiate)

  foc_geom_a_b = lambda geom, a, b: first_order_conditions(
      geom, a, b, f, g, tau_a, tau_b, lse_mode, precondition_fun)

  # Carries pullback onto original inputs, here geom, a and b.
  _, pull_geom_a_b = jax.vjp(foc_geom_a_b, geom, a, b)
  g_geom, g_a, g_b = pull_geom_a_b(vjp_gr)

  # First gradients are for init_a, init_b, momentum, threshold and
  # norm_errors (see jax.custom_vjp definition): those are set to None.
  return None, None, None, None, None, g_geom, g_a, g_b


# Sets threshold, norm_errors, geom, a and b to be differentiable, as those are
# non static. Only differentiability w.r.t. geom, a and b will be used.
_sinkhorn_iterations_implicit = functools.partial(
    jax.custom_vjp, nondiff_argnums=range(16))(
        _sinkhorn_iterations)
_sinkhorn_iterations_implicit.defvjp(_sinkhorn_iterations_taped,
                                     _sinkhorn_iterations_implicit_bwd)


def marginal_error(
    geom: geometry.Geometry,
    a: jnp.ndarray,
    b: jnp.ndarray,
    tau_a: float,
    tau_b: float,
    f_u: jnp.ndarray,
    g_v: jnp.ndarray,
    norm_error: int,
    lse_mode: bool) -> jnp.ndarray:
  """Conputes marginal error, the stopping criterion used to terminate Sinkhorn.

  Args:
    geom: a Geometry object.
    a: jnp.ndarray<float>[num_a,] or jnp.ndarray<float>[batch,num_a] weights.
    b: jnp.ndarray<float>[num_b,] or jnp.ndarray<float>[batch,num_b] weights.
    tau_a: float, ratio lam/(lam+eps) between KL divergence regularizer to first
      marginal and itself + epsilon regularizer used in the unbalanced
      formulation.
    tau_b: float, ratio lam/(lam+eps) between KL divergence regularizer to first
      marginal and itself + epsilon regularizer used in the unbalanced
      formulation.
    f_u: jnp.ndarray, potential or scaling
    g_v: jnp.ndarray, potential or scaling
    norm_error: int, p-norm used to compute error.
    lse_mode: True if log-sum-exp operations, False if kernel vector producs.

  Returns:
    a positive number quantifying how far from convergence the algorithm stands.

  """
  if tau_a == 1.0 and tau_b == 1.0:
    err = geom.error(f_u, g_v, b, 0, norm_error, lse_mode)
  else:
    # In the unbalanced case, we compute the norm of the gradient.
    # the gradient is equal to the marginal of the current plan minus
    # the gradient of < z, rho_z(exp^(-h/rho_z) -1> where z is either a or b
    # and h is either f or g. Note this is equal to z if rho_z → inf, which
    # is the case when tau_z → 1.0
    if lse_mode:
      grad_a = grad_of_marginal_fit(a, f_u, tau_a, geom.epsilon)
      grad_b = grad_of_marginal_fit(b, g_v, tau_b, geom.epsilon)
    else:
      grad_a = grad_of_marginal_fit(a, geom.potential_from_scaling(f_u), tau_a,
                                    geom.epsilon)
      grad_b = grad_of_marginal_fit(b, geom.potential_from_scaling(g_v), tau_b,
                                    geom.epsilon)
    err = geom.error(f_u, g_v, grad_a, 1, norm_error, lse_mode)
    err += geom.error(f_u, g_v, grad_b, 0, norm_error, lse_mode)
  return err


def ent_reg_cost(
    geom: geometry.Geometry,
    a: jnp.ndarray,
    b: jnp.ndarray,
    tau_a: float,
    tau_b: float,
    f: jnp.ndarray,
    g: jnp.ndarray,
    lse_mode: bool) -> jnp.ndarray:
  r"""Computes objective of regularized OT given dual solutions ``f``, ``g``.

  The objective is evaluated for dual solution ``f`` and ``g``, using inputs
  ``geom``, ``a`` and ``b``, in addition to parameters ``tau_a``, ``tau_b``.
  Situations where ``a`` or ``b`` have zero coordinates are reflected in
  minus infinity entries in their corresponding dual potentials. To avoid NaN
  that may result when multiplying 0's by infinity values, ``jnp.where`` is
  used to cancel these contributions.

  Args:
    geom: a Geometry object.
    a: jnp.ndarray<float>[num_a,] or jnp.ndarray<float>[batch,num_a] weights.
    b: jnp.ndarray<float>[num_b,] or jnp.ndarray<float>[batch,num_b] weights.
    tau_a: float, ratio lam/(lam+eps) between KL divergence regularizer to first
      marginal and itself + epsilon regularizer used in the unbalanced
      formulation.
    tau_b: float, ratio lam/(lam+eps) between KL divergence regularizer to first
      marginal and itself + epsilon regularizer used in the unbalanced
      formulation.
    f: jnp.ndarray, potential
    g: jnp.ndarray, potential
    lse_mode: bool, whether to compute total mass in lse or kernel mode.

  Returns:
    a float, the regularized transport cost.
  """
  supp_a = a > 0
  supp_b = b > 0
  if tau_a == 1.0:
    div_a = jnp.sum(
        jnp.where(supp_a, a * (f - geom.potential_from_scaling(a)), 0.0))
  else:
    rho_a = geom.epsilon * (tau_a / (1 - tau_a))
    div_a = -jnp.sum(
        jnp.where(supp_a,
                  a * phi_star(-(f - geom.potential_from_scaling(a)), rho_a),
                  0.0))

  if tau_b == 1.0:
    div_b = jnp.sum(
        jnp.where(supp_b, b * (g - geom.potential_from_scaling(b)), 0.0))
  else:
    rho_b = geom.epsilon * (tau_b / (1 - tau_b))
    div_b = -jnp.sum(
        jnp.where(supp_b,
                  b * phi_star(-(g - geom.potential_from_scaling(b)), rho_b),
                  0.0))

  # Using https://arxiv.org/pdf/1910.12958.pdf (24)
  if lse_mode:
    total_sum = jnp.sum(geom.marginal_from_potentials(f, g))
  else:
    total_sum = jnp.sum(
        geom.marginal_from_scalings(
            geom.scaling_from_potential(f), geom.scaling_from_potential(g)))
  return div_a + div_b + geom.epsilon * (jnp.sum(a) * jnp.sum(b) - total_sum)


def grad_of_marginal_fit(c, h, tau, epsilon):
  """Computes grad of terms linked to marginals in objective.

  Computes gradient w.r.t. f ( or g) of terms in
  https://arxiv.org/pdf/1910.12958.pdf, left-hand-side of Eq. 15
  (terms involving phi_star)

  Args:
    c: jnp.ndarray, first target marginal (either a or b in practice)
    h: jnp.ndarray, potential (either f or g in practice)
    tau: float, strength (in ]0,1]) of regularizer w.r.t. marginal
    epsilon: regularization

  Returns:
    a vector of the same size as c or h
  """
  if tau == 1.0:
    return c
  else:
    rho = epsilon * tau / (1 - tau)
    return jnp.where(c > 0, c * derivative_phi_star(-h, rho), 0.0)


def phi_star(h: jnp.ndarray, rho: float) -> jnp.ndarray:
  """Legendre transform of KL, https://arxiv.org/pdf/1910.12958.pdf p.9."""
  return rho * (jnp.exp(h / rho) - 1)


def derivative_phi_star(f: jnp.ndarray, rho: float) -> jnp.ndarray:
  """Derivative of Legendre transform of KL, see phi_star."""
  return jnp.exp(f / rho)


def second_derivative_phi_star(f: jnp.ndarray, rho: float) -> jnp.ndarray:
  """Second Derivative of Legendre transform of KL, see phi_star."""
  return jnp.exp(f / rho) / rho


def diag_jacobian_of_marginal_fit(c, h, tau, epsilon, derivative):
  """Computes grad of terms linked to marginals in objective.

  Computes second derivative w.r.t. f ( or g) of terms in
  https://arxiv.org/pdf/1910.12958.pdf, left-hand-side of Eq. 32
  (terms involving phi_star)

  Args:
    c: jnp.ndarray, first target marginal (either a or b in practice)
    h: jnp.ndarray, potential (either f or g in practice)
    tau: float, strength (in ]0,1]) of regularizer w.r.t. marginal
    epsilon: regularization
    derivative: Callable

  Returns:
    a vector of the same size as c or h
  """
  if tau == 1.0:
    return 0
  else:
    rho = epsilon * tau / (1 - tau)
    # here no minus sign because we are taking derivative w.r.t -h
    return jnp.where(c > 0,
                     c * second_derivative_phi_star(-h, rho) * derivative(
                         c * derivative_phi_star(-h, rho)),
                     0.0)


def get_transport_functions(geom, lse_mode):
  """Instantiates useful functions from geometry depending on lse_mode."""
  if lse_mode:
    marginal_a = lambda f, g: geom.marginal_from_potentials(f, g, 1)
    marginal_b = lambda f, g: geom.marginal_from_potentials(f, g, 0)
    app_transport = geom.apply_transport_from_potentials
  else:
    marginal_a = lambda f, g: geom.marginal_from_scalings(
        geom.scaling_from_potential(f), geom.scaling_from_potential(g), 1)
    marginal_b = lambda f, g: geom.marginal_from_scalings(
        geom.scaling_from_potential(f), geom.scaling_from_potential(g), 0)
    app_transport = lambda f, g, z, axis: geom.apply_transport_from_scalings(
        geom.scaling_from_potential(f), geom.scaling_from_potential(g), z, axis)
  return marginal_a, marginal_b, app_transport


def solve_implicit_system(
    gr: Tuple[np.ndarray],
    geom: geometry.Geometry,
    a: np.ndarray,
    b: np.ndarray,
    f: np.ndarray,
    g: np.ndarray,
    tau_a: float,
    tau_b: float,
    lse_mode: bool,
    precondition_fun: Callable[[float], float],
    implicit_solver_fun: Callable,
    ridge_kernel: float,
    ridge_identity: float,
    symmetric: bool):
  r"""Applies minus inverse of [hessian of ``reg_ot_cost`` w.r.t ``f``, ``g``].

  This function is used to carry out implicit differentiation of ``sinkhorn``
  outputs, notably optimal potentials ``f`` and ``g``. That differentiation
  requires solving a linear system, using (and inverting) the Jacobian of
  (preconditioned) first-order conditions w.r.t. the reg-OT problem.

  Given a ``precondition_fun``, written here for short as :math:`h`,
  the first order conditions for the dual energy
  :math:`E(K, \epsilon, a, b, f, g) :=- <a,\phi_a^{*}(-f)> + <b, \phi_b^{*}(-g)> - \langle\exp^{f/\epsilon}, K
    \exp^{g/\epsilon}>`

  form the basis of the Sinkhorn algorithm. To differentiate optimal solutions
  to that problem, we exploit the fact that :math:`h(\nabla E = 0)` and
  differentiate that identity to recover variations (Jacobians) of optimal
  solutions :math:`f^\star, g^\star$` as a function of changes in the inputs.
  The Jacobian of :math:`h(\nabla_{f,g} E = 0)` is a linear operator which, if
  it were to be instantiated as a matrix, would be of size
  :math:`(n+m) \times (n+m)`. When :math:`h` is the identity, that matrix is the
  Hessian of :math:`E`, is symmetric and negative-definite
  (:math:`E` is concave) and is structured as :math:`[A, B; B^T, D]`. More
  generally, for other functions :math:`h`, the Jacobian of these preconditioned
  first order conditions is no longer symmetric (except if ``a==b``), and
  has now a structure as :math:`[A, B; C, D]`. That system can
  be still inverted more generic solvers. By default, :math:`h = \epsilon \log`,
  as proposed in https://arxiv.org/pdf/2002.03229.pdf.

  In both cases :math:`A` and :math:`D` are diagonal matrices, equal to the row and
  column marginals respectively, multiplied by the derivatives of :math:`h`
  evaluated at those marginals, corrected (if handling the unbalanced case)
  by the second derivative of the part of the objective that ties potentials to
  the marginals (terms in ``phi_star``). When :math:`h` is the identity,
  :math:`B` and :math:`B^T` are equal respectively to the OT matrix and its
  transpose, i.e. :math:`n \times m` and :math:`m \times n` matrices.
  When :math:`h` is not the identity, :math:`B` (resp. :math:`C`) is equal
  to the OT matrix (resp. its transpose), rescaled on the left by the
  application elementwise of :math:`h'` to the row (respectively column)
  marginal sum of the transport.

  Note that we take great care in not instantiatiating these transport
  matrices, to rely instead on calls to the ``app_transport`` method from the
  ``Geometry`` object ``geom`` (which will either use potentials or scalings,
  depending on ``lse_mode``)

  The Jacobian's diagonal + off-diagonal blocks structure allows to exploit
  Schur complements. Depending on the sizes involved, it is better to
  instantiate the Schur complement of the first or of the second diagonal block.

  In either case, the Schur complement is rank deficient, with a 0 eigenvalue
  for the vector of ones in the balanced case, which is why we add a ridge on
  that subspace to enforce solutions have zero sum.

  The Schur complement can also be rank deficient if two lines or columns of T
  are colinear. This will typically happen it two rows or columns of the cost
  or kernel matrix are numerically close. To avoid this, we add a more global
  ``ridge_identity * z`` regularizer to achieve better conditioning.

  These linear systems are solved using the user defined ``implicit_solver_fun``,
  which is set by default to ``cg``. When the system is symmetric (as detected
  by the corresponding flag ``symmetric``), ``cg`` is applied directly. When it
  is not, normal equations are used (i.e. the Schur complement is multiplied by
  its transpose before solving the system).

  Args:
    gr: 2-uple, (vector of size ``n``, vector of size ``m``).
    geom: Geometry object.
    a: marginal.
    b: marginal.
    f: potential, w.r.t marginal a.
    g: potential, w.r.t marginal b.
    tau_a: float, ratio lam/(lam+eps), ratio of regularizers, first marginal.
    tau_b: float, ratio lam/(lam+eps), ratio of regularizers, second marginal.
    lse_mode: bool, log-sum-exp mode if True, kernel else.
    precondition_fun: preconditioning function to stabilize FOC implicit system.
    implicit_solver_fun: Callable, should return (solution, ...)
    ridge_kernel: promotes zero-sum solutions. only used if tau_a = tau_b = 1.0
    ridge_identity: handles rank deficient transport matrices (this happens
      typically when rows/cols in cost/kernel matrices are colinear, or,
      equivalently when two points from either measure are close).
    symmetric: flag used to figure out whether the linear system solved in the
      implicit function theorem is symmetric or not. This happens when either
      ``a == b`` or the precondition_fun is the identity. False by default, and,
      at the moment, needs to be set manually by the user in the more favorable
      case where the system is guaranteed to be symmetric.

  Returns:
    A tuple of two vectors, of the same size as ``gr``.
  """
  marginal_a, marginal_b, app_transport = get_transport_functions(
      geom, lse_mode)
  # elementwise vmap apply of derivative of precondition_fun. No vmapping
  # can be problematic here.
  derivative = jax.vmap(jax.grad(precondition_fun))

  n, m = geom.shape

  vjp_fg = lambda z: app_transport(
      f, g, z * derivative(marginal_b(f, g)), axis=1) / geom.epsilon
  vjp_gf = lambda z: app_transport(
      f, g, z * derivative(marginal_a(f, g)), axis=0) / geom.epsilon

  if not symmetric:
    vjp_fgt = lambda z: app_transport(
        f, g, z, axis=0) * derivative(marginal_b(f, g)) / geom.epsilon
    vjp_gft = lambda z: app_transport(
        f, g, z, axis=1) * derivative(marginal_a(f, g)) / geom.epsilon

  diag_hess_a = (
      marginal_a(f, g) * derivative(marginal_a(f, g)) / geom.epsilon +
      diag_jacobian_of_marginal_fit(a, f, tau_a, geom.epsilon, derivative))
  diag_hess_b = (
      marginal_b(f, g) * derivative(marginal_b(f, g)) / geom.epsilon +
      diag_jacobian_of_marginal_fit(b, g, tau_b, geom.epsilon, derivative))

  n, m = geom.shape
  # Remove ridge on kernel space if problem is balanced.
  ridge_kernel = jnp.where(tau_a == 1.0 and tau_b == 1.0, ridge_kernel, 0.0)

  # Forks on using Schur complement of either A or D, depending on size.
  if n > m:  #  if n is bigger, run m x m linear system.
    inv_vjp_ff = lambda z: z / diag_hess_a
    vjp_gg = lambda z: z * diag_hess_b
    schur_ = lambda z: vjp_gg(z) - vjp_gf(inv_vjp_ff(vjp_fg(z)))
    g0, g1 = vjp_gf(inv_vjp_ff(gr[0])), gr[1]

    if symmetric:
      schur = lambda z: (schur_(z) + ridge_kernel * jnp.sum(z)
                         + ridge_identity * z)
    else:
      schur_t = lambda z: vjp_gg(z) - vjp_fgt(inv_vjp_ff(vjp_gft(z)))
      g0, g1 = schur_t(g0), schur_t(g1)
      schur = lambda z: (
          schur_t(schur_(z)) + ridge_kernel * jnp.sum(z) + ridge_identity * z)

    sch_f = implicit_solver_fun(schur, g0)[0]
    sch_g = implicit_solver_fun(schur, g1)[0]
    vjp_gr_f = inv_vjp_ff(gr[0] + vjp_fg(sch_f) - vjp_fg(sch_g))
    vjp_gr_g = -sch_f + sch_g
  else:
    vjp_ff = lambda z: z * diag_hess_a
    inv_vjp_gg = lambda z: z / diag_hess_b
    schur_ = lambda z: vjp_ff(z) - vjp_fg(inv_vjp_gg(vjp_gf(z)))
    g0, g1 = vjp_fg(inv_vjp_gg(gr[1])), gr[0]

    if symmetric:
      schur = lambda z: (schur_(z) + ridge_kernel * jnp.sum(z)
                         + ridge_identity * z)
    else:
      schur_t = lambda z: vjp_ff(z) - vjp_gft(inv_vjp_gg(vjp_fgt(z)))
      g0, g1 = schur_t(g0), schur_t(g1)
      schur = lambda z: (schur_t(schur_(z)) + ridge_kernel * jnp.sum(z)
                         + ridge_identity * z)

    sch_g = implicit_solver_fun(schur, g0)[0]
    sch_f = implicit_solver_fun(schur, g1)[0]
    vjp_gr_g = inv_vjp_gg(gr[1] + vjp_gf(sch_g) - vjp_gf(sch_f))
    vjp_gr_f = -sch_g + sch_f

  return jnp.concatenate((-vjp_gr_f, -vjp_gr_g))


def first_order_conditions(
    geom: geometry.Geometry,
    a: jnp.ndarray,
    b: jnp.ndarray,
    f: jnp.ndarray,
    g: jnp.ndarray,
    tau_a: float,
    tau_b: float,
    lse_mode: bool,
    precondition_fun: Optional[Callable[[float], float]]):
  r"""Computes vector of first order conditions for the reg-OT problem.

  The output of this vector should be close to zero at optimality.
  Upon completion of the Sinkhorn forward pass, its norm (using the norm
  parameter defined using ``norm_error``) should be below the threshold
  parameter.

  This error will be itself assumed to be close to zero when using implicit
  differentiation.

  Args:
    geom: a geometry object
    a: jnp.ndarray, first marginal
    b: jnp.ndarray, second marginal
    f: jnp.ndarray, first potential
    g: jnp.ndarray, second potential
    tau_a: float, ratio lam/(lam+eps), ratio of regularizers, first marginal
    tau_b: float, ratio lam/(lam+eps), ratio of regularizers, second marginal
    lse_mode: bool
    precondition_fun: elementwise to apply on both sides of FOC.

  Returns:
    a jnp.ndarray of size (size of ``n + m``) quantifying deviation to
    optimality for variables ``f`` and ``g``.
  """
  marginal_a, marginal_b, _ = get_transport_functions(geom, lse_mode)
  grad_a = grad_of_marginal_fit(a, f, tau_a, geom.epsilon)
  grad_b = grad_of_marginal_fit(b, g, tau_b, geom.epsilon)
  return jnp.concatenate((
      jnp.where(a > 0,
                precondition_fun(marginal_a(f, g)) - precondition_fun(grad_a),
                0.0),
      jnp.where(b > 0,
                precondition_fun(marginal_b(f, g)) - precondition_fun(grad_b),
                0.0)))
